package com.starbuds.server.service.media.impl;

import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.media.BattleStatusEnum;
import com.starbuds.server.common.constant.media.BattleTypeEnum;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.db.LiveBattleDO;
import com.starbuds.server.common.pojo.media.BattleInviteDTO;
import com.starbuds.server.common.pojo.message.im.live.LiveBattleCancelMsg;
import com.starbuds.server.common.pojo.message.im.live.LiveBattleDecisionMsg;
import com.starbuds.server.common.pojo.message.im.live.LiveBattleInviteMsg;
import com.starbuds.server.common.pojo.mq.QueueBattleInviteTimeoutMsg;
import com.starbuds.server.common.service.media.BattleService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.AnchorDAO;
import com.starbuds.server.service.media.dao.LiveBattleDAO;
import com.starbuds.server.service.media.provider.BattleProvider;
import com.starbuds.server.service.media.provider.QiniuLiveProvider;
import com.starbuds.server.service.media.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;

/**
 * PK 服务实现
 *
 * @author Lay
 * @date 2020/4/20
 */
@Service(version = "1.0.0")
@Slf4j
public class BattleServiceImpl implements BattleService {

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Resource
    AnchorDAO anchorDAO;

    @Resource
    LiveBattleDAO liveBattleDAO;

    @Resource
    BattleProvider battleProvider;

    @Resource
    RocketProducer rocketProducer;

    @Resource
    QiniuLiveProvider qiniuLiveProvider;

    /**
     * 邀请 pk
     *
     * @param battleInviteDTO
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public String invite(BattleInviteDTO battleInviteDTO) {
        Long curTime = System.currentTimeMillis();
        Long userId = battleInviteDTO.getUserId();
        Long destUserId = battleInviteDTO.getDestUserId();
        String battleId = BattleProvider.getBattleId(userId, destUserId, curTime);

        try {
            // 修改主播状态
            if (anchorDAO.updateBattling(userId, BoolIntVal.True, BoolIntVal.False) == 0) {
                throw new RuntimeException("主播正在 PK 中");
            }

            // 修改目标主播状态
            if (anchorDAO.updateBattling(destUserId, BoolIntVal.True, BoolIntVal.False) == 0) {
                throw new RuntimeException("目标主播正在 PK 中");
            }

            // 创建记录
            LiveBattleDO liveBattle = new LiveBattleDO();
            liveBattle.setBattleId(battleId);
            liveBattle.setUserId(userId);
            liveBattle.setDestUserId(destUserId);
            liveBattle.setBattleStatus(BattleStatusEnum.Ringing.getValue());
            liveBattle.setBattleType(BattleTypeEnum.Common.getValue());
            liveBattle.setBattleInviteTime(curTime);
            liveBattleDAO.insert(liveBattle);

            // 发送信令消息
            LiveBattleInviteMsg inviteMsg = new LiveBattleInviteMsg();
            inviteMsg.setBattleId(battleId);
            inviteMsg.setInviter(userService.getCard(userId));
            imService.sendLiveRoomMsg(ImTargetType.USER, userId, destUserId, inviteMsg);

            // 超时检查
            QueueBattleInviteTimeoutMsg timeoutMsg = new QueueBattleInviteTimeoutMsg();
            timeoutMsg.setBattleId(battleId);
            timeoutMsg.setUserId(userId);
            timeoutMsg.setDestUserId(destUserId);
            rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.BATTLE_INVITE_TIMEOUT, timeoutMsg.serialize(), curTime + TimeConstants.SECOND_30_MILLS);

            return battleId;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("邀请 pk 失败", e);
        }

        return null;
    }

    /**
     * 获取 pk 状态
     *
     * @param battleId
     * @return
     */
    @Override
    public BattleStatusEnum getBattleStatus(String battleId) {
        Integer battleStatus = liveBattleDAO.selectStatus(battleId);
        if (battleStatus != null) {
            return BattleStatusEnum.toEnum(battleStatus);
        }
        return null;
    }

    /**
     * 获取 PK 记录
     *
     * @param battleId
     * @return
     */
    @Override
    public LiveBattleDO getBattle(String battleId) {
        return liveBattleDAO.selectById(battleId);
    }

    /**
     * 取消邀请
     *
     * @param battleId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean inviteCancel(String battleId) {
        LiveBattleDO liveBattle = liveBattleDAO.selectById(battleId);
        if (liveBattle == null) {
            return false;
        }

        if (liveBattleDAO.updateBattleStatus(battleId, BattleStatusEnum.Canceled.getValue(), BattleStatusEnum.Ringing.getValue()) > 0) {
            // 恢复主播状态
            anchorDAO.updateBattling(liveBattle.getUserId(), BoolIntVal.False, BoolIntVal.True);
            anchorDAO.updateBattling(liveBattle.getDestUserId(), BoolIntVal.False, BoolIntVal.True);

            // 发消息给对方
            LiveBattleCancelMsg cancelMsg = new LiveBattleCancelMsg();
            cancelMsg.setBattleId(battleId);
            imService.sendLiveRoomMsg(ImTargetType.USER, null, liveBattle.getDestUserId(), cancelMsg);

            return true;
        }
        return false;
    }

    /**
     * 邀请决策
     *
     * @param battleId
     * @param decision
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean inviteDecide(String battleId, int decision) {
        LiveBattleDO liveBattle = liveBattleDAO.selectById(battleId);
        if (liveBattle == null) {
            return false;
        }

        Long curTime = System.currentTimeMillis();

        // 决定消息
        LiveBattleDecisionMsg decisionMsg = new LiveBattleDecisionMsg();
        decisionMsg.setBattleId(battleId);
        decisionMsg.setDecision(decision);

        if (BoolIntVal.True == decision) {
            if (battleProvider.startBattle(liveBattle)) {
                // 更新决定时间
                liveBattleDAO.updateDecideTime(battleId, curTime);

                // 发送决定
                decisionMsg.setRoomToken(getRoomToken(battleId, liveBattle.getUserId()));
                imService.sendLiveRoomMsg(ImTargetType.USER, null, liveBattle.getUserId(), decisionMsg);
                return true;
            }
        } else {
            // 拒绝 PK
            if (liveBattleDAO.updateBattleStatus(battleId, BattleStatusEnum.Rejected.getValue(), BattleStatusEnum.Ringing.getValue()) > 0) {
                // 更新决定时间
                liveBattleDAO.updateDecideTime(battleId, curTime);

                // 恢复主播状态
                anchorDAO.updateBattling(liveBattle.getUserId(), BoolIntVal.False, BoolIntVal.True);
                anchorDAO.updateBattling(liveBattle.getDestUserId(), BoolIntVal.False, BoolIntVal.True);

                // 发送决定
                imService.sendLiveRoomMsg(ImTargetType.USER, null, liveBattle.getUserId(), decisionMsg);
            }
            return true;
        }
        return false;
    }

    /**
     * 加入 pk 房间
     *
     * @param battleId
     * @param userId
     * @return
     */
    @Override
    public boolean joinBattleRoom(String battleId, Long userId) {
        return battleProvider.joinBattleRoom(battleId, userId);
    }

    /**
     * 获取房间令牌
     *
     * @param battleId
     * @param userId
     * @return
     */
    @Override
    public String getRoomToken(String battleId, Long userId) {
        return qiniuLiveProvider.getUserRoomToken(battleId, userId, TimeConstants.DAY_MILLS);
    }

    /**
     * 获取 PK 状态
     *
     * @param battleId
     * @return
     */
    @Override
    public Integer getBattlingStatus(String battleId) {
        return battleProvider.getBattlingStatus(battleId);
    }

    /**
     * 关闭 PK
     *
     * @param battleId
     * @return
     */
    @Override
    public boolean closeBattle(String battleId) {
        return battleProvider.battleOver(battleId);
    }

    /**
     * 认输
     *
     * @param battleId
     * @param userId
     * @return
     */
    @Override
    public boolean surrenderBattle(String battleId, Long userId) {
        return battleProvider.surrender(battleId, userId);
    }
}
