package com.starbuds.server.service.media.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.cache.RedisLiveKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.finance.ScoreTypeEnum;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.media.LiveStatusEnum;
import com.starbuds.server.common.constant.media.LiveStreamNotify;
import com.starbuds.server.common.constant.media.StreamStatusEnum;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.pojo.api.PageList;
import com.starbuds.server.common.pojo.api.RecordList;
import com.starbuds.server.common.pojo.data.NearUserDTO;
import com.starbuds.server.common.pojo.data.NearbyUser;
import com.starbuds.server.common.pojo.db.*;
import com.starbuds.server.common.pojo.media.*;
import com.starbuds.server.common.pojo.message.im.live.LiveProductShowMsg;
import com.starbuds.server.common.pojo.message.im.live.LiveUserShareMsg;
import com.starbuds.server.common.pojo.mq.*;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.data.DataService;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.data.UserSearchService;
import com.starbuds.server.common.service.media.LiveService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.user.AnchorService;
import com.starbuds.server.common.service.user.ChannelService;
import com.starbuds.server.service.media.dao.*;
import com.starbuds.server.service.media.provider.*;
import com.starbuds.server.service.media.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * 直播服务实现
 *
 * @author Lay
 * @date 2019/12/23
 */
@Service(version = "1.0.0")
@Slf4j
public class LiveServiceImpl implements LiveService {

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    AnchorService anchorService;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    UserSearchService userSearchService;

    @Reference(version = "1.0.0")
    DataService dataService;

    @Reference(version = "1.0.0")
    ChannelService channelService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Resource
    QiniuLiveProvider qiniuLiveProvider;

    @Resource
    LiveHeatProvider liveHeatProvider;

    @Resource
    LivePopularityProvider livePopularityProvider;

    @Resource
    ChatMatchProvider chatMatchProvider;

    @Resource
    LiveHisDAO liveHisDAO;

    @Resource
    LiveStreamHisDAO liveStreamHisDAO;

    @Resource
    AnchorDAO anchorDAO;

    @Resource
    LiveUserJoinHisDAO liveUserJoinHisDAO;

    @Resource
    LiveTopSuggestHisDAO liveTopSuggestHisDAO;

    @Resource
    LiveShareHisDAO liveShareHisDAO;

    @Resource
    LiveCategoryDAO liveCategoryDAO;

    @Resource
    RocketProducer rocketProducer;

    @Resource
    ProviderBuilder providerBuilder;

    /**
     * 推流过期时间
     */
    private static final int livePushUrlExpireSeconds = 24 * 3600;

    /**
     * 拉流过期时间
     */
    private static final int livePullUrlExpireSeconds = 60;

    /**
     * 获取推流地址
     *
     * @param userId
     * @param liveHisId
     * @return
     */
    @Override
    public String getStreamPushUrl(Long userId, Long liveHisId) {
        Long streamHisId = identityService.getId();
        return qiniuLiveProvider.getPushUrl(userId, liveHisId, streamHisId, livePushUrlExpireSeconds);
    }

    /**
     * 获取拉流地址
     *
     * @param userId
     * @return
     */
    @Override
    public LiveStreamPullUrl getStreamPullUrl(Long userId) {
        return qiniuLiveProvider.getPullUrl(userId, livePullUrlExpireSeconds);
    }

    /**
     * 开始直播
     *
     * @param userId
     * @param liveTitle
     * @param liveCover
     * @param categoryId
     * @param areaId
     * @param longitude
     * @param latitude
     * @param channelId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public LiveStartVO startLive(Long userId, String liveTitle, String liveCover, Long categoryId, Integer areaId, Double longitude, Double latitude, Long channelId) {
        return startLive(userId, liveTitle, liveCover, categoryId, areaId, longitude, latitude, channelId, null);
    }

    /**
     * 开始直播
     *
     * @param userId
     * @param liveTitle
     * @param liveCover
     * @param categoryId
     * @param areaId
     * @param longitude
     * @param latitude
     * @param channelId
     * @param productIds
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public LiveStartVO startLive(Long userId, String liveTitle, String liveCover, Long categoryId, Integer areaId, Double longitude, Double latitude, Long channelId, String productIds) {
        try {
            Long hisId = identityService.getId();
            long curTime = System.currentTimeMillis();

            if (StringUtils.isBlank(liveCover)) {
                liveCover = anchorService.getLiveCover(userId);
            }

            // 创建直播记录
            LiveHisDO liveHisDO = new LiveHisDO();
            liveHisDO.setHisId(hisId);
            liveHisDO.setUserId(userId);
            liveHisDO.setCategoryId(categoryId);
            liveHisDO.setLiveStatus(LiveStatusEnum.Living.getValue());
            liveHisDO.setLiveTitle(liveTitle);
            liveHisDO.setLiveCover(liveCover);
            liveHisDO.setLiveStartTime(curTime);
            liveHisDO.setHisCreateTime(curTime);
            liveHisDO.setProductIds(productIds);
            liveHisDAO.insert(liveHisDO);

            // 更新直播间信息
            AnchorDO anchorDO = new AnchorDO();
            anchorDO.setUserId(userId);
            anchorDO.setLiveStatus(LiveStatusEnum.Living.getValue());
            anchorDO.setLiveTitle(liveTitle);
            anchorDO.setLiveCover(liveCover);
            anchorDO.setCategoryId(categoryId);
            anchorDO.setAreaId(areaId);
            anchorDO.setLongitude(longitude);
            anchorDO.setLatitude(latitude);
            anchorDO.setLastLiveTime(curTime);
            if (!anchorService.editAnchor(anchorDO)) {
                throw new Exception("修改主播信息失败");
            }

            LiveStartVO liveStartVO = new LiveStartVO();
            liveStartVO.setPushUrl(getStreamPushUrl(userId, hisId));
            liveStartVO.setLiveHisId(hisId);

            // 初始化热度值
            liveHeatProvider.initLiveHeatAsync(userId);

            // 初始化人气
            livePopularityProvider.init(userId);

            // 初始化魅力值、星币
            redisService.hashPut(RedisLiveKeys.LIVE_ANCHOR_CHARM, String.valueOf(userId), "0");
            redisService.hashPut(RedisLiveKeys.LIVE_ANCHOR_COIN, String.valueOf(userId), "0");

            // 发送开播消息
            QueueLiveStartMsg queueLiveStartMsg = new QueueLiveStartMsg();
            queueLiveStartMsg.setAnchorUserId(userId);
            queueLiveStartMsg.setHisId(hisId);
            queueLiveStartMsg.setChannelId(channelId);
            queueLiveStartMsg.setTimestamp(curTime);
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.LIVE_START, queueLiveStartMsg.serialize());

            // 将主播移除单聊匹配池
            chatMatchProvider.removeMatchPoolAnchor(userId);

            if (StringUtils.isNotBlank(productIds)) {
                // 缓存直播商品
                redisService.hashPut(RedisLiveKeys.LIVE_PRODUCTS, String.valueOf(userId), productIds);

                // 重置展示商品
                redisService.hashPut(RedisLiveKeys.LIVE_SHOW_PRODUCT, String.valueOf(userId), String.valueOf(0));
            }

            return liveStartVO;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("开播失败", e);
        }
        return null;
    }

    /**
     * 下播接口
     *
     * @param hisId
     * @param isTimeout
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public LiveDataVO stopLive(Long hisId, boolean isTimeout) {
        log.info("下播 {} 超时 {}", hisId, isTimeout);
        LiveHisDO liveHisDO = liveHisDAO.selectById(hisId);
        if (liveHisDO == null) {
            return null;
        }

        LiveDataVO liveDataVO = new LiveDataVO();

        try {
            long curTime = System.currentTimeMillis();
            Long userId = liveHisDO.getUserId();

            // 计算直播时长和有效时长
            Long liveDuration = curTime - liveHisDO.getLiveStartTime();

            // 判断有效时长是否已计算
            if (isTimeout && liveHisDO.getLiveEffectiveDuration() == null) {
                try {
                    List<LiveStreamHisDO> liveStreamHisDOList = liveStreamHisDAO.selectByLiveHisId(liveHisDO.getHisId());
                    Long liveEffectiveDuration = 0L;
                    if (CollectionUtil.isNotEmpty(liveStreamHisDOList)) {
                        for (LiveStreamHisDO liveStreamHis : liveStreamHisDOList) {
                            liveEffectiveDuration += liveStreamHis.getStreamStopTime() - liveStreamHis.getStreamStartTime();
                        }
                    }
                    liveHisDO.setLiveEffectiveDuration(liveEffectiveDuration * 1000L);
                } catch (Exception e) {
                    log.error("直播有效时长计算失败了");
                }
            }

            // 更新直播记录
            Long liveHeatMax = liveHeatProvider.getLiveHeatMax(userId);
            Long livePopularityMax = livePopularityProvider.getMax(userId);
            Long liveCharmIncome = Optional.ofNullable(redisService.hashGet(RedisLiveKeys.LIVE_ANCHOR_CHARM, String.valueOf(userId))).map(t -> Convert.toLong(t)).orElse(0L);
            Long liveCoinIncome = Optional.ofNullable(redisService.hashGet(RedisLiveKeys.LIVE_ANCHOR_COIN, String.valueOf(userId))).map(t -> Convert.toLong(t)).orElse(0L);
            Integer giftNum = Optional.ofNullable(redisService.hashGet(RedisLiveKeys.LIVE_ANCHOR_GIFT, String.valueOf(userId))).map(t -> Convert.toInt(t)).orElse(0);
            liveHisDO.setLiveStopTime(curTime);
            liveHisDO.setLiveDuration(liveDuration);
            liveHisDO.setLiveHeatMax(liveHeatMax);
            liveHisDO.setLiveMaxPopularity(livePopularityMax);
            liveHisDO.setLiveStatus(LiveStatusEnum.Closed.getValue());
            liveHisDO.setLiveIsTimeout(isTimeout ? BoolIntVal.True : BoolIntVal.False);
            liveHisDO.setLiveCoinIncome(liveCoinIncome);
            liveHisDO.setGiftNum(giftNum);
            liveHisDO.setHisUpdateTime(curTime);
            liveHisDAO.update(liveHisDO);

            // 更新主播信息
            AnchorDO anchorDO = new AnchorDO();
            anchorDO.setUserId(userId);
            anchorDO.setLiveStatus(LiveStatusEnum.Closed.getValue());
            if (!anchorService.editAnchor(anchorDO)) {
                throw new Exception("更新主播信息失败");
            }

            // 移除热度值
            liveHeatProvider.removeLiveHeat(userId);

            // 移除人气
            livePopularityProvider.remove(userId);

            // 移除魅力值、星币收入
            redisService.hashDel(RedisLiveKeys.LIVE_ANCHOR_CHARM, String.valueOf(userId), "0");
            redisService.hashDel(RedisLiveKeys.LIVE_ANCHOR_COIN, String.valueOf(userId), "0");
            redisService.hashDel(RedisLiveKeys.LIVE_ANCHOR_GIFT, String.valueOf(userId), "0");

            // 移除置顶推荐
            removeTopSuggest(userId);

            // 移除直播商品
            redisService.hashDel(RedisLiveKeys.LIVE_PRODUCTS, String.valueOf(userId));

            // 重置展示商品
            redisService.hashDel(RedisLiveKeys.LIVE_SHOW_PRODUCT, String.valueOf(userId));

            // 移除直播游戏
            redisService.hashDel(RedisLiveKeys.LIVE_GAME, String.valueOf(userId));

            // 发送下播消息
            QueueLiveStopMsg queueLiveStopMsg = new QueueLiveStopMsg();
            queueLiveStopMsg.setAnchorUserId(liveHisDO.getUserId());
            queueLiveStopMsg.setHisId(hisId);
            queueLiveStopMsg.setLiveDuration(liveDuration);
            queueLiveStopMsg.setLiveHeatMax(liveHeatMax);
            queueLiveStopMsg.setLivePopularityMax(livePopularityMax);
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.LIVE_STOP, queueLiveStopMsg.serialize());

            liveDataVO.setLiveHisId(hisId);
            liveDataVO.setAnchorUserId(liveHisDO.getUserId());
            liveDataVO.setLiveHeatMax(liveHeatMax);
            liveDataVO.setLivePopularityMax(livePopularityMax);
            liveDataVO.setLiveDuration(liveDuration);

            // 魅力值增量
            Map<Integer, Long> scoreIncomeMap = new HashMap<>();
            scoreIncomeMap.put(ScoreTypeEnum.CHARM.getValue(), liveCharmIncome);
            liveDataVO.setScoreIncome(scoreIncomeMap);

            Map<Integer, Long> coinIncomeMap = new HashMap<>();
            coinIncomeMap.put(WalletTypeEnum.STAR_COIN.getValue(), liveCoinIncome);
            liveDataVO.setWalletIncome(coinIncomeMap);

            // 将主播加回单聊匹配池
            chatMatchProvider.addMatchPoolAnchor(userId);

            return liveDataVO;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("下播失败", e);
        }
        return null;
    }

    /**
     * 推流通知处理
     *
     * @param liveStreamNotify
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public boolean handleStreamNotify(LiveStreamNotify liveStreamNotify) {
        try {
            LiveStreamHisDO liveStreamHisDO;
            if (StreamStatusEnum.Pushing == liveStreamNotify.getStreamStatus()) {
                // 推流开始
                liveStreamHisDO = new LiveStreamHisDO();
                liveStreamHisDO.setHisId(liveStreamNotify.getStreamHisId());
                liveStreamHisDO.setLiveHisId(liveStreamNotify.getLiveHisId());
                liveStreamHisDO.setStreamStatus(StreamStatusEnum.Pushing.getValue());
                liveStreamHisDO.setStreamStartTime(liveStreamNotify.getTimesteamp());
                liveStreamHisDO.setHisIpAddress(liveStreamNotify.getIp());
                liveStreamHisDO.setHisCreateTime(System.currentTimeMillis());
                liveStreamHisDAO.insert(liveStreamHisDO);
            } else {
                // 推流结束
                liveStreamHisDO = liveStreamHisDAO.selectById(liveStreamNotify.getStreamHisId());
                if (liveStreamHisDO == null) {
                    return false;
                }

                // 保存直播回放
                liveStreamHisDO.setStreamStatus(StreamStatusEnum.Closed.getValue());
                liveStreamHisDO.setStreamStopTime(liveStreamNotify.getTimesteamp());
                liveStreamHisDAO.update(liveStreamHisDO);

                // 判断直播记录状态
                LiveHisDO liveHisDO = liveHisDAO.selectById(liveStreamNotify.getLiveHisId());
                if (liveHisDO.getLiveStopTime() != null) {
                    // 已关播 计算直播有效时长
                    List<LiveStreamHisDO> liveStreamHisDOList = liveStreamHisDAO.selectByLiveHisId(liveStreamNotify.getLiveHisId());
                    Long liveEffectiveDuration = 0L;
                    if (CollectionUtil.isNotEmpty(liveStreamHisDOList)) {
                        for (LiveStreamHisDO liveStreamHis : liveStreamHisDOList) {
                            liveEffectiveDuration += liveStreamHis.getStreamStopTime() - liveStreamHis.getStreamStartTime();
                        }
                    }
                    liveHisDAO.updateEffectiveDurationById(liveStreamNotify.getLiveHisId(), liveEffectiveDuration * 1000L);
                } else {
                    // 丢出一条超时检查消息 5 分钟后检查直播状态
                    QueueLiveTimeoutMsg liveTimeoutMsg = new QueueLiveTimeoutMsg();
                    liveTimeoutMsg.setLiveHisId(liveStreamNotify.getLiveHisId());
                    liveTimeoutMsg.setStreamHisId(liveStreamNotify.getStreamHisId());
                    rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.LIVE_TIMEOUT, liveTimeoutMsg.serialize(), System.currentTimeMillis() + 60 * 1000L);
                }
            }

            // 更新主播状态
            AnchorDO anchorDO = new AnchorDO();
            anchorDO.setUserId(liveStreamNotify.getUserId());
            anchorDO.setStreamStatus(liveStreamNotify.getStreamStatus().getValue());
            if (!anchorService.editAnchor(anchorDO)) {
                throw new Exception("主播推流状态更新失败");
            }

            if (StreamStatusEnum.Pushing == liveStreamNotify.getStreamStatus()) {
                QueueLiveStreamPushMsg queueLiveStreamPushMsg = new QueueLiveStreamPushMsg();
                queueLiveStreamPushMsg.setAnchorUserId(liveStreamNotify.getUserId());
                queueLiveStreamPushMsg.setLiveHisId(liveStreamNotify.getLiveHisId());
                queueLiveStreamPushMsg.setStreamHisId(liveStreamNotify.getStreamHisId());
                rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.STREAM_PUSH, queueLiveStreamPushMsg.serialize());
            } else {
                QueueLiveStreamCloseMsg queueLiveStreamCloseMsg = new QueueLiveStreamCloseMsg();
                queueLiveStreamCloseMsg.setAnchorUserId(liveStreamNotify.getUserId());
                queueLiveStreamCloseMsg.setLiveHisId(liveStreamNotify.getLiveHisId());
                queueLiveStreamCloseMsg.setStreamHisId(liveStreamNotify.getStreamHisId());
                queueLiveStreamCloseMsg.setStartTime(liveStreamHisDO.getStreamStartTime());
                queueLiveStreamCloseMsg.setStopTime(liveStreamHisDO.getStreamStopTime());
                rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.STREAM_CLOSE, queueLiveStreamCloseMsg.serialize());
            }

            return true;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("推流通知处理失败", e);
        }
        return false;
    }

    /**
     * 直播记录是否存在
     *
     * @param hisId
     * @param userId
     * @return
     */
    @Override
    public boolean isLiveHisExists(Long hisId, Long userId) {
        return liveHisDAO.countById(hisId, userId) > 0;
    }

    /**
     * 改变直播热度
     *
     * @param userId
     * @param heatValue
     * @param duration
     * @return
     */
    @Override
    public Long changeLiveHeat(Long userId, Long heatValue, Long duration) {
        return liveHeatProvider.changeLiveHeat(userId, heatValue, Optional.ofNullable(duration).orElse(0L));
    }

    /**
     * 获取置顶推荐直播
     *
     * @param topSuggestDTO
     * @return
     */
    @Override
    public List<LiveItemVO> getTopSuggestList(TopSuggestDTO topSuggestDTO) {
        // 坑位总数
        final int POS_TOTAL = 6;

        List<LiveItemVO> list = new ArrayList<>();

        Set<Long> userIds = new HashSet<>();
        Set<Long> tagIds = new HashSet<>();
        Map<Long, Integer> positionMap = new HashMap<>();
        Map<Long, Long> userTagMap = new HashMap<>();
        Map<Long, Long> userHeatMap = new HashMap<>();

        // 查询指定推荐渠道
        Long channelId = channelService.getTopSuggestChannelId(topSuggestDTO.getChannelId());

        // 查询置顶直播
        List<String> cachedList = redisService.batchGet(redisService.getKeys(RedisLiveKeys.getLiveTopSuggestPattern(channelId)));
        if (CollectionUtil.isNotEmpty(cachedList)) {
            for (String item : cachedList) {
                TopSuggest topSuggest = JSON.parseObject(item, TopSuggest.class);
                if (topSuggest.getTagId() != null) {
                    tagIds.add(topSuggest.getTagId());
                    userTagMap.put(topSuggest.getUserId(), topSuggest.getTagId());
                }

                userIds.add(topSuggest.getUserId());
                positionMap.put(topSuggest.getUserId(), topSuggest.getPosition());
            }
        }

        // 热门补坑位
        int posNum = POS_TOTAL - userIds.size();
        if (posNum > 0) {
            List<LiveHeat> liveHeats = liveHeatProvider.getTopN(0, POS_TOTAL);

            if (CollectionUtil.isNotEmpty(liveHeats)) {
                liveHeats.sort(Comparator.comparing(LiveHeat::getHeatValue).reversed());
                int index = 1;
                for (LiveHeat liveHeat : liveHeats) {
                    // 热度预存 如果这里已经有热度了 后面可以少取一次
                    userHeatMap.put(liveHeat.getUserId(), liveHeat.getHeatValue());

                    // 重复跳过
                    if (userIds.contains(liveHeat.getUserId())) {
                        continue;
                    }

                    if (posNum >= index) {
                        // 补个坑
                        userIds.add(liveHeat.getUserId());
                        // 设置坑位
                        positionMap.put(liveHeat.getUserId(), POS_TOTAL + index);
                        // 补了个坑
                        index++;
                    }
                }
            }
        }

        // 执行处理
        if (CollectionUtil.isNotEmpty(userIds)) {
            List<LiveItemVO> itemVOList = anchorService.getLiveItemsByIds(new ArrayList<>(userIds));
            if (CollectionUtil.isNotEmpty(itemVOList)) {
                // 数据标签准备
                Map<Long, DataLiveTagDO> tagMap = new HashMap<>();
                if (CollectionUtil.isNotEmpty(tagIds)) {
                    List<DataLiveTagDO> tagList = dataService.getLiveTagByIds(new ArrayList<>(tagIds));
                    if (CollectionUtil.isNotEmpty(tagList)) {
                        tagMap = tagList.stream().collect(Collectors.toMap(DataLiveTagDO::getTagId, Function.identity()));
                    }
                }

                // 写入值
                for (LiveItemVO liveItemVO : itemVOList) {
                    Long userId = liveItemVO.getUserId();
                    // 写入热度
                    Long heatValue = Optional.ofNullable(userHeatMap.get(userId)).orElse(liveHeatProvider.getLiveHeat(userId));
                    liveItemVO.setLiveHeat(heatValue);
                    // 判断标签
                    if (userTagMap.containsKey(userId)) {
                        DataLiveTagDO tag = tagMap.get(userTagMap.get(userId));
                        if (tag != null) {
                            liveItemVO.setTagId(tag.getTagId());
                            liveItemVO.setTagIcon(tag.getTagIcon());
                            liveItemVO.setTagName(tag.getTagName());
                        }
                    }
                    // 写入位置
                    liveItemVO.setPosition(positionMap.get(userId));

                    // 渠道id
                    liveItemVO.setChannelId(topSuggestDTO.getChannelId());
                    liveItemVO.setChannelName(channelService.getChannelName(topSuggestDTO.getChannelId()));
                }

                // 排序
                itemVOList.sort(Comparator.comparing(LiveItemVO::getPosition));
                list.addAll(itemVOList);
            }
        }

        return list;
    }

    /**
     * 获取直播列表
     *
     * @param pageIndex
     * @param pageSize
     * @param categoryId
     * @param liveStatus
     * @return
     */
    @Override
    public PageList<LiveItemVO> getList(int pageIndex, int pageSize, Long categoryId, LiveStatusEnum liveStatus) {
        PageList<LiveItemVO> pageList = new PageList<>(pageIndex, pageSize);
        List<LiveItemVO> liveList = anchorDAO.selectLiveItems(pageList.getPageCursor(), pageSize, categoryId, liveStatus.getValue());

        for (Iterator<LiveItemVO> iterator = liveList.iterator(); iterator.hasNext(); ) {
            LiveItemVO liveItem = iterator.next();
            liveItem.setLiveHeat(liveHeatProvider.getLiveHeat(liveItem.getUserId()));
        }

        pageList.setList(liveList);
        pageList.setTotal(anchorDAO.countLiveItems(categoryId, liveStatus.getValue()));

        return pageList;
    }

    /**
     * 获取热门列表
     *
     * @param lastId
     * @param pageSize
     * @return
     */
    @Override
    public RecordList<LiveItemVO, Integer> getHeatList(int lastId, int pageSize) {
        RecordList<LiveItemVO, Integer> heatList = new RecordList<>(pageSize);
        heatList.setLastId(0);

        // 查询热门
        List<LiveHeat> liveHeats = liveHeatProvider.getTopN(lastId, pageSize);

        if (CollectionUtil.isNotEmpty(liveHeats)) {
            Map<Long, Long> userHeatMap = liveHeats.stream().collect(Collectors.toMap(LiveHeat::getUserId, LiveHeat::getHeatValue));
            List<Long> userIds = liveHeats.stream().map(LiveHeat::getUserId).collect(Collectors.toList());

            // 查询主播信息
            List<LiveItemVO> itemVOList = anchorService.getLiveItemsByIds(userIds);

            // 写入热度
            for (LiveItemVO liveItemVO : itemVOList) {
                liveItemVO.setLiveHeat(userHeatMap.get(liveItemVO.getUserId()));
            }

            // 存入结果
            itemVOList.sort(Comparator.comparingLong(LiveItemVO::getLiveHeat).reversed());
            heatList.setList(itemVOList);

            // 设置下一页标记
            if (itemVOList.size() == pageSize) {
                heatList.setLastId(lastId + pageSize);
            }
        }

        return heatList;
    }

    /**
     * 获取关注直播列表
     *
     * @param pageIndex
     * @param pageSize
     * @param userId
     * @param liveStatus
     * @return
     */
    @Override
    public PageList<LiveItemVO> getFollowList(int pageIndex, int pageSize, Long userId, LiveStatusEnum liveStatus) {
        PageList<LiveItemVO> pageList = new PageList<>(pageIndex, pageSize);
        List<LiveItemVO> liveList = liveHisDAO.selectAllFollowItems(pageList.getPageCursor(), pageSize, userId, liveStatus.getValue());

        for (Iterator<LiveItemVO> iterator = liveList.iterator(); iterator.hasNext(); ) {
            LiveItemVO liveItem = iterator.next();
            liveItem.setLiveHeat(liveHeatProvider.getLiveHeat(liveItem.getUserId()));
        }

        pageList.setList(liveList);
        pageList.setTotal(liveHisDAO.countFollowItems(userId, liveStatus.getValue()));

        return pageList;
    }

    /**
     * 获取附近的直播列表
     *
     * @param pageIndex
     * @param pageSize
     * @param userId
     * @param liveStatus
     * @param areaId
     * @param latitude
     * @param longitude
     * @param distance
     * @param userSex
     * @return
     */
    @Override
    public PageList<LiveItemVO> getNearbyList(int pageIndex, int pageSize, Long userId, LiveStatusEnum liveStatus, Integer areaId, Double latitude, Double longitude, Double distance, Integer userSex) {

        NearUserDTO userDTO = new NearUserDTO();
        userDTO.setAreaId(areaId);
        userDTO.setUserLatitude(latitude);
        userDTO.setUserLongitude(longitude);
        userDTO.setUserSex(userSex);
        userDTO.setIsAnchor(BoolIntVal.True);
        userDTO.setLiveStatusEnum(liveStatus);
        userDTO.setDistance(distance);
        userDTO.setSingleChatEnable(null);
        userDTO.setIsOnline(null);
        userDTO.setUserAge(null);
        userDTO.setSingleChatCategoryId(null);
        userDTO.setPageIndex(pageIndex);
        userDTO.setPageSize(pageSize);

        PageList<NearbyUser> pageListNearbyUser = userSearchService.getNearUserPageList(userDTO);
        PageList<LiveItemVO> pageList = new PageList<>(pageIndex, pageSize);
        pageList.setTotal(pageListNearbyUser.getTotal());

        if (CollectionUtil.isNotEmpty(pageListNearbyUser.getList())) {
            // 准备 userId
            List<Long> userIds = pageListNearbyUser.getList().stream().map(t -> t.getUserId()).collect(Collectors.toList());

            // 查询直播的其他东西
            List<LiveItemVO> liveItemList = new ArrayList<>();
            Map<Long, LiveItemVO> itemVOMap = anchorDAO.selectLiveItemsByUserList(userIds).stream().collect(Collectors.toMap(t -> t.getUserId(), Function.identity()));

            for (Iterator<NearbyUser> iterator = pageListNearbyUser.getList().iterator(); iterator.hasNext(); ) {
                NearbyUser nearbyUser = iterator.next();
                if (itemVOMap.containsKey(nearbyUser.getUserId())) {
                    LiveItemVO liveItemVO = itemVOMap.get(nearbyUser.getUserId());
                    liveItemVO.setDistance(nearbyUser.getDistance());
                    // 热度
                    liveItemVO.setLiveHeat(liveHeatProvider.getLiveHeat(nearbyUser.getUserId()));
                    liveItemList.add(liveItemVO);
                }
            }

            pageList.setList(liveItemList);
        }

        return pageList;
    }

    /**
     * 禁播流
     *
     * @param userId
     * @param stopTime
     * @return
     */
    @Override
    public boolean disableStream(Long userId, long stopTime) {
        return qiniuLiveProvider.disableStream(userId, stopTime / 1000);
    }

    /**
     * 取消禁播流
     *
     * @param userId
     * @return
     */
    @Override
    public boolean enableStream(Long userId) {
        return qiniuLiveProvider.disableStream(userId, 0);
    }

    /**
     * 获取直播间在线人数
     *
     * @param userId 用户id
     * @return 在线人数
     */
    @Override
    public long getOnlineCount(Long userId) {
        return Optional.ofNullable(livePopularityProvider.get(userId)).orElse(0L);
    }

    /**
     * 获取直播间热度
     *
     * @param userId 用户id
     * @return 在线人数
     */
    @Override
    public long getLiveHeat(Long userId) {
        return Optional.ofNullable(liveHeatProvider.getLiveHeat(userId)).orElse(0L);
    }

    /**
     * 获取用户进房/出房次数
     *
     * @param userId
     * @param anchorUserId
     * @param joinType
     * @param timestamp
     * @return
     */
    @Override
    public int getLiveUserJoinDayBefore(Long userId, Long anchorUserId, Integer joinType, Long timestamp) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(timestamp);
        calendar.set(Calendar.HOUR_OF_DAY, 0);
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);
        long beginTime = calendar.getTimeInMillis();
        return liveUserJoinHisDAO.countByTypeWithTime(userId, anchorUserId, joinType, beginTime, timestamp);
    }

    /**
     * 获取主播当日开播次数
     *
     * @param userId
     * @param timestamp
     * @return
     */
    @Override
    public int getLiveTimesDayBefore(Long userId, Long timestamp) {
        Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(timestamp);
        calendar.set(Calendar.HOUR_OF_DAY, 0);
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);
        long beginTime = calendar.getTimeInMillis();
        return liveHisDAO.countByTime(userId, beginTime, timestamp);
    }

    /**
     * 获取直播记录列表
     *
     * @param pageIndex
     * @param pageSize
     * @param userId
     * @param yearMonth
     * @return
     */
    @Override
    public PageList<LiveHisVO> getLiveHisList(int pageIndex, int pageSize, Long userId, Integer yearMonth) {
        Long beginTime = null, endTime = null;
        if (yearMonth != null) {
            try {
                SimpleDateFormat sdf = new SimpleDateFormat("yyyyMM");
                Calendar calendar = Calendar.getInstance();
                calendar.setTime(sdf.parse(String.valueOf(yearMonth)));
                calendar.set(Calendar.DAY_OF_MONTH, 1);
                calendar.set(Calendar.HOUR_OF_DAY, 0);
                calendar.set(Calendar.MINUTE, 0);
                calendar.set(Calendar.SECOND, 0);
                calendar.set(Calendar.MILLISECOND, 0);
                beginTime = calendar.getTimeInMillis();
                calendar.add(Calendar.MONTH, 1);
                calendar.add(Calendar.DATE, -1);
                calendar.set(Calendar.HOUR_OF_DAY, 23);
                calendar.set(Calendar.MINUTE, 59);
                calendar.set(Calendar.SECOND, 59);
                calendar.set(Calendar.MILLISECOND, 999);
                endTime = calendar.getTimeInMillis();
            } catch (Exception e) {
                log.error("年月转换失败");
            }
        }

        PageList<LiveHisVO> pageList = new PageList<>(pageIndex, pageSize);
        pageList.setList(liveHisDAO.selectAll(pageList.getPageCursor(), pageSize, userId, beginTime, endTime));
        pageList.setTotal(liveHisDAO.countAll(userId, beginTime, endTime));
        return pageList;
    }

    /**
     * 设置置顶推荐
     *
     * @param channelId
     * @param userId
     * @param position
     * @param tagId
     * @param duration
     * @param adminId
     * @return
     */
    @Override
    public boolean setTopSuggest(Long channelId, Long userId, int position, Long tagId, Long duration, Long adminId) {
        try {
            // 获取指定渠道的推荐列表
            List<String> values = redisService.batchGet(redisService.getKeys(RedisLiveKeys.getLiveTopSuggestPattern(channelId)));
            if (CollectionUtil.isNotEmpty(values)) {
                List<String> keysToDel = new ArrayList<>();
                for (String item : values) {
                    TopSuggest topSuggest = JSON.parseObject(item, TopSuggest.class);
                    if (userId.equals(topSuggest.getUserId())) {
                        // 加入待移除
                        keysToDel.add(RedisLiveKeys.getLiveTopSuggest(channelId, topSuggest.getPosition()));
                    }
                }
                if (CollectionUtil.isNotEmpty(keysToDel)) {
                    redisService.delKeys(keysToDel);
                }
            }

            // 设置置顶推荐
            TopSuggest topSuggest = new TopSuggest();
            topSuggest.setChannelId(channelId);
            topSuggest.setUserId(userId);
            topSuggest.setTagId(tagId);
            topSuggest.setPosition(position);
            if (duration != null && duration > 0) {
                redisService.set(RedisLiveKeys.getLiveTopSuggest(channelId, position), JSON.toJSONString(topSuggest), duration, TimeUnit.MILLISECONDS);
            } else {
                redisService.set(RedisLiveKeys.getLiveTopSuggest(channelId, position), JSON.toJSONString(topSuggest));
            }

            // 写入记录
            LiveTopSuggestHisDO liveTopSuggestHis = new LiveTopSuggestHisDO();
            liveTopSuggestHis.setHisId(identityService.getId());
            liveTopSuggestHis.setChannelId(channelId);
            liveTopSuggestHis.setUserId(userId);
            liveTopSuggestHis.setPosition(position);
            liveTopSuggestHis.setTagId(tagId);
            liveTopSuggestHis.setDuration(duration);
            liveTopSuggestHis.setAdminId(adminId);
            liveTopSuggestHis.setHisCreateTime(System.currentTimeMillis());
            liveTopSuggestHisDAO.insert(liveTopSuggestHis);

            return true;
        } catch (Exception e) {
            log.error("置顶推荐设置失败", e);
        }
        return false;
    }

    /**
     * 移除置顶推荐
     *
     * @param channelId
     * @param position
     * @return
     */
    @Override
    public boolean removeTopSuggest(Long channelId, int position) {
        try {
            redisService.delKey(RedisLiveKeys.getLiveTopSuggest(channelId, position));
            return true;
        } catch (Exception e) {
            log.error("置顶推荐移除失败", e);
        }
        return false;
    }

    /**
     * 移动置顶推荐位置
     *
     * @param channelId
     * @param curPosition
     * @param destPosition
     * @param isExchange   是否交换
     * @return
     */
    @Override
    public boolean moveTopSuggest(Long channelId, int curPosition, int destPosition, boolean isExchange) {
        try {
            // 无法自己跟自己换
            if (curPosition == destPosition) {
                return false;
            }

            String curKey = RedisLiveKeys.getLiveTopSuggest(channelId, curPosition);
            String destKey = RedisLiveKeys.getLiveTopSuggest(channelId, destPosition);

            String curItem = redisService.get(curKey);
            String destItem = redisService.get(destKey);

            // 先判断当前位置是不是空的
            if (StringUtils.isBlank(curItem)) {
                return false;
            }

            // 无论如何先覆盖掉
            TopSuggest curSuggest = JSON.parseObject(curItem, TopSuggest.class);
            curSuggest.setPosition(destPosition);
            redisService.set(destKey, JSON.toJSONString(curSuggest));

            // 判断是不是交换
            if (isExchange && StringUtils.isNotBlank(destItem)) {
                TopSuggest destSuggest = JSON.parseObject(destItem, TopSuggest.class);
                destSuggest.setPosition(curPosition);
                redisService.set(curKey, JSON.toJSONString(destSuggest));
            } else {
                // 直接干掉
                redisService.delKey(curKey);
            }

            return true;
        } catch (Exception e) {
            log.error("置顶推荐位置移动失败");
        }
        return false;
    }

    /**
     * 移除置顶推荐
     *
     * @param userId
     * @return
     */
    @Override
    public boolean removeTopSuggest(Long userId) {
        try {
            // 获取指定渠道的推荐列表
            List<String> values = redisService.batchGet(redisService.getKeys(RedisLiveKeys.getLiveTopSuggestPattern()));
            if (CollectionUtil.isNotEmpty(values)) {
                List<String> keysToDel = new ArrayList<>();
                for (String item : values) {
                    TopSuggest topSuggest = JSON.parseObject(item, TopSuggest.class);
                    if (userId.equals(topSuggest.getUserId())) {
                        // 加入待移除
                        keysToDel.add(RedisLiveKeys.getLiveTopSuggest(topSuggest.getChannelId(), topSuggest.getPosition()));
                    }
                }
                if (CollectionUtil.isNotEmpty(keysToDel)) {
                    redisService.delKeys(keysToDel);
                }
            }
            return true;
        } catch (Exception e) {
            log.error("置顶推荐移除失败", e);
        }
        return false;
    }

    /**
     * 分享直播间
     *
     * @param anchorUserId
     * @param userId
     * @param channelId
     * @return
     */
    @Override
    public boolean shareLive(Long anchorUserId, Long userId, Long channelId) {
        // 判断1分钟内是否分享过
        LiveShareHisDO lastHis = liveShareHisDAO.selectLast(anchorUserId, userId);
        if (lastHis != null) {
            if (System.currentTimeMillis() - lastHis.getHisCreateTime() < TimeConstants.MIN_MILLS) {
                return false;
            }
        }

        try {
            LiveShareHisDO liveShareHis = new LiveShareHisDO();
            liveShareHis.setHisId(identityService.getId());
            liveShareHis.setAnchorUserId(anchorUserId);
            liveShareHis.setUserId(userId);
            liveShareHis.setHisCreateTime(System.currentTimeMillis());
            liveShareHisDAO.insert(liveShareHis);

            // 发送消息
            LiveUserShareMsg msg = new LiveUserShareMsg();
            imService.sendLiveRoomMsg(ImTargetType.GROUP, userId, anchorUserId, msg);

            // 加入统计
            statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.LIVE_SHARE_TIMES, BigDecimal.ONE);

            return true;
        } catch (Exception e) {
            log.error("直播间分享记录创建失败", e);
        }
        return true;
    }

    @Override
    public Long getLivePopularityMax(Long userId) {
        return livePopularityProvider.get(userId);
    }

    /**
     * 直播类目是否允许售卖商品
     *
     * @param categoryId
     * @return
     */
    @Override
    public boolean isCategorySellEnabled(Long categoryId) {
        return Optional.ofNullable(liveCategoryDAO.selectSellEnable(categoryId)).map(t -> BoolIntVal.True == t).orElse(false);
    }

    /**
     * 获取直播售卖信息
     *
     * @param userId
     * @return
     */
    @Override
    public LiveSellVO getLiveSell(Long userId) {
        LiveSellVO liveSellVO = new LiveSellVO();
        liveSellVO.setProductIds(Optional.ofNullable(redisService.hashGet(RedisLiveKeys.LIVE_PRODUCTS, String.valueOf(userId))).map(Object::toString).orElse(null));
        liveSellVO.setShowProductId(Optional.ofNullable(redisService.hashGet(RedisLiveKeys.LIVE_SHOW_PRODUCT, String.valueOf(userId))).map(Convert::toLong).orElse(0L));
        return liveSellVO;
    }

    /**
     * 设置直播展示商品
     *
     * @param userId
     * @param productId
     * @return
     */
    @Override
    public boolean setShowProductId(Long userId, Long productId) {
        try {
            redisService.hashPut(RedisLiveKeys.LIVE_SHOW_PRODUCT, String.valueOf(userId), String.valueOf(productId));
        } catch (Exception e) {
            log.error("设置直播展示商品失败", e);
            return false;
        }

        // 发送展示商品改变消息
        LiveProductShowMsg liveProductShowMsg = new LiveProductShowMsg();
        liveProductShowMsg.setProductId(productId);
        imService.sendLiveRoomMsg(ImTargetType.GROUP, null, userId, liveProductShowMsg);

        return true;
    }

    /**
     * 判断用户是否在频道中
     *
     * @param channelName
     * @param userId
     * @return
     */
    @Override
    public boolean getRemoteUserExistChannel(String channelName, Long userId, String appVer) {
        return providerBuilder.provider(appVer).isInChannel(channelName, userId);
    }

}
