package com.starbuds.server.service.media.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.cache.RedisCacheKeys;
import com.starbuds.server.common.pojo.media.LiveRankItem;
import com.starbuds.server.common.pojo.media.LiveRankVO;
import com.starbuds.server.common.pojo.user.UserCardVO;
import com.starbuds.server.common.service.core.CacheService;
import com.starbuds.server.common.service.media.RankService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.LiveRankDAO;
import com.starbuds.server.service.media.provider.RankProvider;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.util.CollectionUtils;

import java.util.*;

/**
 * 排行榜服务
 *
 * @author Lay
 * @date 2020/1/10
 */
@Service(version = "1.0.0")
public class RankServiceImpl implements RankService {

    @Autowired
    RankProvider rankProvider;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    CacheService cacheService;

    @Autowired
    LiveRankDAO liveRankDAO;

    /**
     * 增加直播排行榜数据
     *
     * @param anchorUserId
     * @param userId
     * @param amount
     */
    @Override
    public void addLiveRankAmount(Long anchorUserId, Long userId, Long amount) {
        rankProvider.addLiveRankAmount(anchorUserId, userId, amount);
    }

    /**
     * 获取直播日榜
     *
     * @param anchorUserId
     * @param start
     * @param end
     * @return
     */
    @Override
    public List<LiveRankItem> getLiveRankDay(Long anchorUserId, long start, long end) {
        String[] times = rankProvider.getRankTimes();
        String redisKey = RedisCacheKeys.getLiveRankDayKey(anchorUserId, times[0]);

        return rankProvider.getLiveRankItems(redisKey, start, end);
    }

    /**
     * 获取直播周榜
     *
     * @param anchorUserId
     * @param start
     * @param end
     * @return
     */
    @Override
    public List<LiveRankItem> getLiveRankWeek(Long anchorUserId, long start, long end) {
        String[] times = rankProvider.getRankTimes();
        String redisKey = RedisCacheKeys.getLiveRankWeekKey(anchorUserId, times[1]);

        return rankProvider.getLiveRankItems(redisKey, start, end);
    }

    /**
     * 获取直播月榜
     *
     * @param anchorUserId
     * @param start
     * @param end
     * @return
     */
    @Override
    public List<LiveRankItem> getLiveRankMonth(Long anchorUserId, long start, long end) {
        String[] times = rankProvider.getRankTimes();
        String redisKey = RedisCacheKeys.getLiveRankMonthKey(anchorUserId, times[2]);

        return rankProvider.getLiveRankItems(redisKey, start, end);
    }

    /**
     * 获取直播总榜
     *
     * @param anchorUserId
     * @param start
     * @param end
     * @return
     */
    @Override
    public List<LiveRankVO> getLiveRankTotal(Long anchorUserId, long start, long end) {
        int pageCursor = Convert.toInt(start);
        int pageSize = Convert.toInt(end - start + 1);
        List<LiveRankVO> list = liveRankDAO.selectItems(anchorUserId, pageCursor, pageSize);
        if (CollectionUtil.isNotEmpty(list)) {
            list.sort(Comparator.comparingLong(LiveRankVO::getAmount).reversed().thenComparingLong(LiveRankVO::getRankUpdateTime));
            // 写入排序号
            int rankNo = pageCursor + 1;
            for (Iterator<LiveRankVO> iterator = list.iterator(); iterator.hasNext(); ) {
                iterator.next().setRankNo(rankNo);
                rankNo++;
            }
            return list;
        }

        return Collections.EMPTY_LIST;
    }

    /**
     * 获取日榜贡献值
     *
     * @param anchorUserId
     * @param userId
     * @return
     */
    @Override
    public LiveRankItem getDayRankItem(Long anchorUserId, Long userId) {
        LiveRankItem item = new LiveRankItem();
        String[] times = rankProvider.getRankTimes();
        String redisKey = RedisCacheKeys.getLiveRankDayKey(anchorUserId, times[0]);
        Integer rankNo = Optional.ofNullable(cacheService.zRevRank(redisKey, String.valueOf(userId))).map(Long::intValue).orElse(-1);
        Long amount = Optional.ofNullable(cacheService.zScore(redisKey, String.valueOf(userId))).map(Convert::toLong).orElse(0L);
        List<UserCardVO> userCardVOList = userService.getNameAndAvatarByIds(Collections.singletonList(userId));
        if (!CollectionUtils.isEmpty(userCardVOList)) {
            UserCardVO userCardVO = userCardVOList.get(0);
            item.setUserName(userCardVO.getUserName());
            item.setUserAvatar(userCardVO.getUserAvatar());
            item.setWealthLevel(userCardVO.getWealthLevel());
        }
        item.setAmount(amount);
        item.setRankNo(rankNo + 1);
        item.setUserId(userId);
        return item;
    }

    /**
     * 获取周榜贡献值
     *
     * @param anchorUserId
     * @param userId
     * @return
     */
    @Override
    public LiveRankItem getWeekRankItem(Long anchorUserId, Long userId) {
        LiveRankItem item = new LiveRankItem();
        String[] times = rankProvider.getRankTimes();
        String redisKey = RedisCacheKeys.getLiveRankWeekKey(anchorUserId, times[1]);
        Integer rankNo = Optional.ofNullable(cacheService.zRevRank(redisKey, String.valueOf(userId))).map(t -> t.intValue()).orElse(-1);
        Long amount = Optional.ofNullable(cacheService.zScore(redisKey, String.valueOf(userId))).map(t -> Convert.toLong(t)).orElse(0L);
        List<UserCardVO> userCardVOList = userService.getNameAndAvatarByIds(Arrays.asList(userId));
        if (!CollectionUtils.isEmpty(userCardVOList)) {
            UserCardVO userCardVO = userCardVOList.get(0);
            item.setUserName(userCardVO.getUserName());
            item.setUserAvatar(userCardVO.getUserAvatar());
            item.setWealthLevel(userCardVO.getWealthLevel());
        }
        item.setAmount(amount);
        item.setRankNo(rankNo + 1);
        item.setUserId(userId);
        return item;
    }

    /**
     * 获取月榜贡献值
     *
     * @param anchorUserId
     * @param userId
     * @return
     */
    @Override
    public LiveRankItem getMonthRankItem(Long anchorUserId, Long userId) {
        LiveRankItem item = new LiveRankItem();
        String[] times = rankProvider.getRankTimes();
        String redisKey = RedisCacheKeys.getLiveRankMonthKey(anchorUserId, times[2]);
        Integer rankNo = Optional.ofNullable(cacheService.zRevRank(redisKey, String.valueOf(userId))).map(t -> t.intValue()).orElse(-1);
        Long amount = Optional.ofNullable(cacheService.zScore(redisKey, String.valueOf(userId))).map(t -> Convert.toLong(t)).orElse(0L);
        List<UserCardVO> userCardVOList = userService.getNameAndAvatarByIds(Arrays.asList(userId));
        if (!CollectionUtils.isEmpty(userCardVOList)) {
            UserCardVO userCardVO = userCardVOList.get(0);
            item.setUserName(userCardVO.getUserName());
            item.setUserAvatar(userCardVO.getUserAvatar());
            item.setWealthLevel(userCardVO.getWealthLevel());
        }
        item.setAmount(amount);
        item.setRankNo(rankNo + 1);
        item.setUserId(userId);
        return item;
    }

    /**
     * 获取排行时间
     *
     * @return
     */
    @Override
    public String[] getRankTimes() {
        return rankProvider.getRankTimes();
    }

}
