package com.starbuds.server.service.media.impl;

import cn.hutool.core.util.RandomUtil;
import cn.hutool.crypto.SecureUtil;
import com.starbuds.server.common.constant.media.GameShareTargetTypeEnum;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.pojo.message.im.game.GameShareMsg;
import com.starbuds.server.common.pojo.user.UserCardVO;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.media.ThirdGameService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.util.Url;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.beans.factory.annotation.Value;

import java.util.concurrent.TimeUnit;

/**
 * 三方游戏服务实现
 *
 * @author Lay
 * @date 2021/6/24
 */
@Service(version = "1.0.0")
public class ThirdGameServiceImpl implements ThirdGameService {

    @Value("${gamebox.channelId}")
    private String channelId;

    @Value("${gamebox.key}")
    private String key;

    @Value("${gamebox.gateWay}")
    private String gateWay;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    ImService imService;

    /**
     * 获取游戏大厅地址
     *
     * @param userId
     * @return
     */
    @Override
    public String getGameHallUrl(Long userId) {
        // 拉取用户信息
        UserCardVO userCard = userService.getCard(userId);
        if (userCard != null) {
            Long curTime = System.currentTimeMillis() / 1000;
            StringBuilder sbUrl = new StringBuilder();
            sbUrl.append(gateWay);
            sbUrl.append("/hall?");
            sbUrl.append(String.format("uid=%d", userId));
            sbUrl.append(String.format("&nick=%s", Url.encodeURIComponent(userCard.getUserName())));
            sbUrl.append(String.format("&avatar=%s", Url.encodeURIComponent(userCard.getUserAvatar())));
            sbUrl.append(String.format("&sex=%d", userCard.getUserSex()));
            sbUrl.append(String.format("&channel=%s", channelId));
            sbUrl.append(String.format("&time=%d", curTime));
            sbUrl.append(String.format("&sign=%s", sign(curTime, userId)));
            return sbUrl.toString();
        }

        return null;
    }

    /**
     * 是否分享限制
     *
     * @param userId
     * @param targetType
     * @param targetId
     * @return
     */
    @Override
    public boolean isShareLimited(Long userId, int targetType, Long targetId) {
        GameShareTargetTypeEnum gameShareTargetType = GameShareTargetTypeEnum.toEnum(targetType);
        if (gameShareTargetType == null) {
            return true;
        }

        if (GameShareTargetTypeEnum.PublicChat == gameShareTargetType) {
            // 公聊大厅限制 1分钟 1次
            String limitKey = String.format("game:share:limit_%d_%d", targetType, userId);
            if (redisService.setIfAbsent(limitKey, String.valueOf(System.currentTimeMillis()), 1, TimeUnit.MINUTES)) {
                return false;
            }
        } else {
            // 15秒一次
            String limitKey = String.format("game:share:limit_%d_%d_%d", targetType, userId, targetId);
            if (redisService.setIfAbsent(limitKey, String.valueOf(System.currentTimeMillis()), 15, TimeUnit.SECONDS)) {
                return false;
            }
        }

        return true;
    }

    /**
     * 分享游戏
     *
     * @param userId
     * @param params
     * @param gameName
     * @param gameIcon
     * @param targetType
     * @param targetId
     * @return
     */
    @Override
    public boolean share(Long userId, String params, String gameName, String gameIcon, int targetType, Long targetId) {
        GameShareTargetTypeEnum gameShareTargetType = GameShareTargetTypeEnum.toEnum(targetType);
        if (gameShareTargetType == null) {
            return false;
        }

        String[] bodyTips = new String[]{
                "游戏我已就位，就等你来开一局",
                "敢不敢和我PK一下，点我进入战场",
                "没你在我不尽兴，和我一起组CP吧！",
                "终于等到你了，看不到你我玩不下去！",
                "么么 我在输液啦！输的什么液，想你的夜！",
                "我有一堆臭毛病，都是想你的臭毛病！"
        };

        // 构造消息
        GameShareMsg gameShareMsg = new GameShareMsg();
        gameShareMsg.setParams(params);
        gameShareMsg.setGameName(gameName);
        gameShareMsg.setGameIcon(gameIcon);
        gameShareMsg.setTitle(String.format("我正在玩%s，快来和我一起玩！", gameName));
        gameShareMsg.setBody(bodyTips[RandomUtil.randomInt(0, bodyTips.length)]);

        // 发送消息
        if (GameShareTargetTypeEnum.PublicChat == gameShareTargetType) {
            // 分享到公聊大厅
            return imService.sendRtcRoomMsg(ImTargetType.ALL, userId, null, null, gameShareMsg);
        } else if (GameShareTargetTypeEnum.Fans == gameShareTargetType) {
            // 分享给粉丝
            return imService.sendCustomImMsg(userId, targetId, gameShareMsg, null);
        }

        return false;
    }

    private String sign(Long time, Long userId) {
        String toSign = String.format("channel=%s&time=%d&uid=%d&key=%s", channelId, time, userId, key);
        return SecureUtil.md5(toSign);
    }
}
