package com.starbuds.server.service.media.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import cn.hutool.core.util.IdUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.starbuds.server.common.constant.cache.RedisChatKeys;
import com.starbuds.server.common.constant.cache.RedisFinanceKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.media.ChatGroupType;
import com.starbuds.server.common.constant.media.ChatStatusEnum;
import com.starbuds.server.common.constant.media.VoiceChatStopReasonEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.constant.statistics.StatUserTypeEnum;
import com.starbuds.server.common.constant.user.VoiceChatStatus;
import com.starbuds.server.common.pojo.api.PageList;
import com.starbuds.server.common.pojo.db.ChatGroupDO;
import com.starbuds.server.common.pojo.db.VoiceChatHisDO;
import com.starbuds.server.common.pojo.media.ChatResponse;
import com.starbuds.server.common.pojo.media.VoiceChatDTO;
import com.starbuds.server.common.pojo.media.VoiceChatRequest;
import com.starbuds.server.common.pojo.media.VoiceChatResponse;
import com.starbuds.server.common.pojo.message.im.chat.VoiceChatCancelMsg;
import com.starbuds.server.common.pojo.message.im.chat.VoiceChatStopMsg;
import com.starbuds.server.common.pojo.mq.*;
import com.starbuds.server.common.pojo.rtc.RtcAuthInfo;
import com.starbuds.server.common.pojo.user.UserMatchResponse;
import com.starbuds.server.common.pojo.user.UserMatchVO;
import com.starbuds.server.common.pojo.user.UserPeiPeiVO;
import com.starbuds.server.common.service.core.IdentityService;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.finance.PayService;
import com.starbuds.server.common.service.finance.WalletService;
import com.starbuds.server.common.service.media.VoiceChatService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.statistics.StatUserService;
import com.starbuds.server.common.service.user.FamilyService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.ChatGroupDAO;
import com.starbuds.server.service.media.dao.VoiceChatHisDAO;
import com.starbuds.server.service.media.provider.AgoraProvider;
import com.starbuds.server.service.media.provider.ProviderBuilder;
import com.starbuds.server.service.media.provider.VoiceChatProvider;
import com.starbuds.server.service.media.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.config.annotation.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * @Author jiangxiaowei
 * @DateTime 2020/11/13 17:42
 * @Description 语音聊天
 */
@Slf4j
@Service(version = "1.0.0")
public class VoiceChatServiceImpl implements VoiceChatService {

    @Resource
    RocketProducer rocketProducer;

    @Resource
    VoiceChatHisDAO voiceChatHisDAO;

    @Resource
    VoiceChatProvider voiceChatProvider;

    @Resource
    ProviderBuilder providerBuilder;

    @Resource
    ChatGroupDAO chatGroupDAO;

    @Reference(version = "1.0.0")
    WalletService walletService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    IdentityService identityService;

    @Reference(version = "1.0.0")
    PayService payService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    FamilyService familyService;

    @Reference(version = "1.0.0")
    private StatUserService statUserService;

    @Reference(version = "1.0.0")
    private StatPlatformService statPlatformService;

    /**
     * 发出语音速配请求
     *
     * @param chatRequest
     * @return
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public VoiceChatResponse sendVoiceMatch(VoiceChatRequest chatRequest) {
        Long userId = chatRequest.getUserId();
        VoiceChatResponse chatResponse = new VoiceChatResponse();
        Long channelId = chatRequest.getChannelId();
        Integer userSex = chatRequest.getUserSex();

        // 统计语音速配请求人数
        statUserService.addUserStat(userId, StatUserTypeEnum.VOICE_CHAT_REQUEST_USER_COUNT.getValue(), BigDecimal.ONE);
        // 统计语音速配按钮点击次数
        statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.VOICE_CHAT_REQUEST_COUNT, BigDecimal.ONE);

        String voiceChatKey = RedisChatKeys.getVoiceChatLock(userId);
        String uuid = IdUtil.simpleUUID();

        try {
            boolean lock = redisService.setIfAbsent(voiceChatKey, uuid, 10, TimeUnit.SECONDS);
            if (lock) {
                long curTime = System.currentTimeMillis();
                // 判断钱包是否足够支持
                Long balance = walletService.getBalance(userId, WalletTypeEnum.STAR_DIAMOND);
                // 如果没有免费时长且钱包不足以支撑1分钟
                if (!hasFreeTime(userId, curTime) && balance < 100L) {
                    chatResponse.setSuccess(false);
                    chatResponse.setMsg("余额不足，请充值");
                    return chatResponse;
                }

                Long hisId = identityService.getId();
                // 创建记录
                VoiceChatHisDO hisDO = new VoiceChatHisDO();
                hisDO.setHisId(hisId);
                hisDO.setUserId(userId);
                hisDO.setHisStatus(ChatStatusEnum.Ringing.getValue());
                hisDO.setHisPriceType(WalletTypeEnum.STAR_DIAMOND.getValue());
                hisDO.setHisPrice(100L);
                hisDO.setHisUpdateTime(curTime);
                hisDO.setHisCreateTime(curTime);
                hisDO.setMatchStartTime(curTime);
                hisDO.setChannelId(channelId);
                hisDO.setHisIsMatch(BoolIntVal.True);
                hisDO.setHisTimeMinutes(0);
                hisDO.setUserSex(userSex);
                voiceChatHisDAO.insert(hisDO);

                // 更新用户聊天状态
                userService.updateVoiceChatStatus(userId, VoiceChatStatus.BUSY);

                chatResponse.setHisId(hisId);
                chatResponse.setSuccess(true);
                chatResponse.setMsg("请求已收到，系统正加速匹配");

                // 发送mq
                QueueVoiceMatchRequestMsg requestMsg = new QueueVoiceMatchRequestMsg();
                requestMsg.setHisId(hisId);
                requestMsg.setUserId(userId);
                requestMsg.setUserSex(userSex);
                requestMsg.setMatchStartTime(curTime);
                rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.VOICE_MATCH_REQUEST, requestMsg.serialize());

                // 发送超时mq
                QueueVoiceMatchTimeoutMsg timeoutMsg = new QueueVoiceMatchTimeoutMsg();
                timeoutMsg.setHisId(hisId);
                timeoutMsg.setMatchStartTime(curTime);
                rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.VOICE_MATCH_TIMEOUT, timeoutMsg.serialize(), curTime + 60 * 1000L);
                // 记录语音速配匹配开始时间
                redisService.set(RedisChatKeys.getVoiceChatMatchStartTime(userId), "" + System.currentTimeMillis(), 2, TimeUnit.MINUTES);
            } else {
                chatResponse.setSuccess(false);
                chatResponse.setMsg("正在速配");
            }
            return chatResponse;
        } catch (Exception e) {
            log.error("发起速配失败", e);
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            chatResponse.setHisId(null);
            chatResponse.setSuccess(false);
            chatResponse.setMsg("发起速配失败");
        } finally {
            if (uuid.equals(redisService.get(voiceChatKey))) {
                redisService.delKey(voiceChatKey);
            }
        }
        return chatResponse;
    }

    /**
     * 判断用户是否空闲状态
     *
     * @param userId
     */
    @Override
    public boolean isUserIdle(Long userId) {
        return voiceChatProvider.isUserIdle(userId);
    }

    /**
     * 获取记录
     *
     * @param hisId
     * @return
     */
    @Override
    public VoiceChatHisDO getChatHis(Long hisId) {
        return voiceChatHisDAO.selectById(hisId);
    }

    /**
     * 获取记录状态
     *
     * @param hisId
     * @return
     */
    @Override
    public Integer getHisStatus(Long hisId) {
        return voiceChatHisDAO.selectHisStatus(hisId);
    }

    /**
     * 同意语音邀请
     *
     * @param hisId
     * @param targetUserId
     * @return
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public ChatResponse startChat(Long hisId, Long targetUserId, String appVer) {
        ChatResponse chatResponse = new ChatResponse();
        log.info("同意速配邀请, hisId {} targetUserId {}", hisId, targetUserId);

        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);
        if (hisDO == null) {
            log.info("未找到记录");
            chatResponse.setSuccess(false);
            chatResponse.setMsg("未找到记录");
            return chatResponse;
        }

        Long userId = hisDO.getUserId();

        // 统计记录语音速配匹配时间
        long matchStartTime = NumberUtils.toLong(redisService.get(RedisChatKeys.getVoiceChatMatchStartTime(userId)), 0);
        if (matchStartTime > 0) {
            long matchDuration = System.currentTimeMillis() - matchStartTime;
            statPlatformService.addPlatformStat(hisDO.getChannelId(), StatPlatformTypeEnum.VOICE_CHAT_MATCH_DURATION, new BigDecimal("" + matchDuration));
        }

        Long hisStartTime = System.currentTimeMillis();
        String chatHandleKey = RedisChatKeys.getVoiceChatHandleLock(hisId);
        String uuid = IdUtil.simpleUUID();

        try {
            // 是否有免费时长
            boolean hasFreeTime = hasFreeTime(userId, hisStartTime);
            Long balance = walletService.getBalance(userId, WalletTypeEnum.STAR_DIAMOND);

            if (hasFreeTime) {
                // 抢占资源
                boolean lock = redisService.setIfAbsent(chatHandleKey, uuid, 10, TimeUnit.SECONDS);
                if (!lock) {
                    chatResponse.setSuccess(false);
                    chatResponse.setMsg("太火了,速配正被抢");
                    return chatResponse;
                }
                redisService.hashPut(RedisFinanceKeys.VOICE_CHAT_CHARGING, Convert.toStr(hisId), "0");
                voiceChatProvider.useFreeTime(userId, hisStartTime);
                // 已计费
                redisService.hashIncreaseLong(RedisFinanceKeys.VOICE_CHAT_CHARGING, Convert.toStr(hisId), 1);
                log.info("使用免费时长,hisId {}, targetUserId {}", hisId, targetUserId);
            } else if (balance < hisDO.getHisPrice()) {
                log.info("余额不足,hisId {}", hisId);
                chatResponse.setSuccess(false);
                chatResponse.setMsg("余额不足");
                return chatResponse;
            } else {
                // 抢占资源
                boolean lock = redisService.setIfAbsent(chatHandleKey, uuid, 10, TimeUnit.SECONDS);
                if (!lock) {
                    chatResponse.setSuccess(false);
                    chatResponse.setMsg("太火了,速配正被抢");
                    return chatResponse;
                }
                redisService.hashPut(RedisFinanceKeys.VOICE_CHAT_CHARGING, Convert.toStr(hisId), "0");
                // 预支付一分钟
                Long payResult = payService.payForVoiceChat(hisId, userId, targetUserId, hisDO.getChannelId(), hisDO.getHisPrice());
                if (payResult == null || payResult < 0) {
                    log.info("扣费失败，或余额不足,hisId {},payResult {}", hisId, payResult);
                    chatResponse.setSuccess(false);
                    chatResponse.setMsg("扣费失败，或对方余额不足");
                    return chatResponse;
                }
                log.info("预支付1分钟成功,hisId {}, targetUserId {}", hisId, targetUserId);
            }

            VoiceChatHisDO chatUpdate = new VoiceChatHisDO();
            chatUpdate.setHisId(hisId);
            chatUpdate.setTargetUserId(targetUserId);
            chatUpdate.setHisTimeMinutes(hasFreeTime ? 1 : 0);
            chatUpdate.setHisStartTime(hisStartTime);
            chatUpdate.setHisUpdateTime(hisStartTime);
            chatUpdate.setHisStatus(ChatStatusEnum.Calling.getValue());
            voiceChatHisDAO.updateById(chatUpdate);

            // 增加匹配次数
            familyService.incrementConnectTimes(targetUserId);

            // 创建群组
            ChatGroupDO chatGroupDO = new ChatGroupDO();
            chatGroupDO.setGroupId(Convert.toStr(hisId));
            chatGroupDO.setOwnerUserId(targetUserId);
            chatGroupDO.setGroupType(ChatGroupType.VOICE_CHAT);
            chatGroupDO.setGroupCreateTime(hisStartTime);
            chatGroupDAO.insert(chatGroupDO);

            // 更新语音状态
            userService.updateVoiceChatStatus(userId, VoiceChatStatus.BUSY);
            userService.updateVoiceChatStatus(targetUserId, VoiceChatStatus.BUSY);

            // 开始录制
            String res = providerBuilder.provider(appVer).startRecord(Convert.toStr(hisId));
            if (StringUtils.isNotBlank(res)) {
                JSONObject jsonObject = JSON.parseObject(res);
                voiceChatHisDAO.updateReplayId(hisId, jsonObject.getString("sid"), jsonObject.getString("resourceId"));
            }

            // 开始mq
            QueueVoiceChatStartMsg startMsg = new QueueVoiceChatStartMsg();
            startMsg.setHisId(hisId);
            startMsg.setUserId(userId);
            startMsg.setChannelId(hisDO.getChannelId());
            startMsg.setTargetUserId(targetUserId);
            startMsg.setHisStartTime(hisStartTime);
            startMsg.setUseFreeTime(hasFreeTime);
            startMsg.setAppVer(appVer);
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.VOICE_CHAT_START, startMsg.serialize());

            long startDeliverTime = hisStartTime + TimeConstants.MIN_MILLS;
            // 计费周期 下一分钟
            QueueVoiceChatChargingTickMsg tickMsg = new QueueVoiceChatChargingTickMsg();
            tickMsg.setHisId(hisId);
            tickMsg.setTargetUserId(targetUserId);
            tickMsg.setTimes(2);
            tickMsg.setUseFreeTime(false);
            tickMsg.setTimestamp(startDeliverTime);
            tickMsg.setAppVer(appVer);
            rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.VOICE_CHAT_CHARGING_TICK, tickMsg.serialize(), startDeliverTime);

            chatResponse.setSuccess(true);
            chatResponse.setMsg("处理成功");

        } catch (Exception e) {
            log.error("出现错误", e);
            chatResponse = new ChatResponse();
            chatResponse.setSuccess(false);
            chatResponse.setMsg("程序出现错误,请反馈运营");
            return chatResponse;
        } finally {
            if (uuid.equals(redisService.get(chatHandleKey))) {
                log.info("用户接通了语音速配 hisId {}, targetUserId {}", hisId, targetUserId);
                redisService.delKey(chatHandleKey);
            }
        }

        return chatResponse;
    }

    /**
     * 拒绝聊天
     *
     * @param hisId
     * @param curUserId
     * @param round
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    @Override
    public ChatResponse rejectChat(Long hisId, Long curUserId, int round) {
        ChatResponse chatResponse = new ChatResponse();
        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);
        if (hisDO == null) {
            log.info("未找到记录");
            chatResponse.setSuccess(false);
            chatResponse.setMsg("未找到记录");
            return chatResponse;
        }

        ChatStatusEnum statusEnum = ChatStatusEnum.toEnum(hisDO.getHisStatus());
        if (statusEnum != ChatStatusEnum.Ringing) {
            log.info("当前速配{}{}不支持拒绝", hisId, statusEnum.getDesc());
            chatResponse.setSuccess(false);
            chatResponse.setMsg("当前速配" + statusEnum.getDesc() + "不支持拒绝");
            return chatResponse;
        }

        Long userId = hisDO.getUserId();
        Long matchStartTime = hisDO.getMatchStartTime();
        long matchShouldEndTime = matchStartTime + TimeConstants.MIN_MILLS;

        // 更新语音状态
        userService.updateVoiceChatStatus(curUserId, VoiceChatStatus.IDLE);

        // 拒绝mq
        QueueVoiceChatRejectMsg rejectMsg = new QueueVoiceChatRejectMsg();
        rejectMsg.setHisId(hisId);
        rejectMsg.setUserId(userId);
        rejectMsg.setTargetUserId(curUserId);
        rejectMsg.setRound(round);
        rejectMsg.setMatchStartTime(matchStartTime);
        rejectMsg.setMatchShouldEndTime(matchShouldEndTime);
        rejectMsg.setExcludedUserIds(voiceChatProvider.getExcludedUserIds(hisId));
        rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.VOICE_CHAT_REJECT, rejectMsg.serialize());

        chatResponse.setSuccess(true);
        chatResponse.setMsg("处理成功");
        return chatResponse;
    }

    /**
     * 停止语音聊天
     *
     * @param hisId
     * @param reasonEnum
     * @param hisStopTime
     * @param times
     * @return
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public ChatResponse stopChat(Long hisId, VoiceChatStopReasonEnum reasonEnum, Long hisStopTime, int times, String appVer) {
        ChatResponse chatResponse = new ChatResponse();
        try {
            VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);

            if (hisDO == null) {
                try {
                    if (times < 3) {
                        Thread.sleep(500L);
                        return stopChat(hisId, reasonEnum, hisStopTime, times++, appVer);
                    }
                } catch (Exception e) {
                    log.info("出现异常", e);
                }
                log.info("未找到记录{}", hisId);
                chatResponse.setSuccess(false);
                chatResponse.setMsg("未找到记录");
                return chatResponse;
            }

            log.info("hisDO {}", JSON.toJSONString(hisDO));

            Long userId = hisDO.getUserId();
            Long targetUserId = hisDO.getTargetUserId();
            ChatStatusEnum statusEnum = ChatStatusEnum.toEnum(hisDO.getHisStatus());
            Long hisStartTime = hisDO.getHisStartTime();
            Long hisPrice = hisDO.getHisPrice();
            Integer hisPriceType = hisDO.getHisPriceType();
            Integer hisTimeMinutes = hisDO.getHisTimeMinutes();
            long curTime = System.currentTimeMillis();

            // 呼叫中
            if (ChatStatusEnum.Ringing == statusEnum) {
                // 直接取消
                int result = voiceChatHisDAO.updateStatusWithMinutes(hisId, ChatStatusEnum.Canceled.getValue(), ChatStatusEnum.Ringing.getValue(), null, hisStopTime, 0, reasonEnum.getValue());
                if (result == 0) {
                    log.info("修改状态失败,hisId {}", hisId);
                    chatResponse.setSuccess(false);
                    chatResponse.setMsg("呼叫中，修改状态失败");
                    return chatResponse;
                }

                // 更新语音状态
                userService.updateVoiceChatStatus(userId, VoiceChatStatus.IDLE);

                log.info("取消消息发给自己，hisId {} , userId {}", hisId, userId);
                // 取消消息发给自己
                VoiceChatCancelMsg cancelMsgToUser = new VoiceChatCancelMsg();
                cancelMsgToUser.setHisId(hisId);
                cancelMsgToUser.setUserId(userId);
                cancelMsgToUser.setTargetUserId(targetUserId);
                cancelMsgToUser.setHisCancelTime(curTime);
                cancelMsgToUser.setMsg("您已取消");
                imService.sendCustomImMsg(null, userId, cancelMsgToUser, null);

                if (targetUserId != null) {
                    log.info("取消消息发给对方，hisId {}, targetUserId {}", hisId, targetUserId);
                    // 更新聊天状态
                    userService.updateVoiceChatStatus(targetUserId, VoiceChatStatus.IDLE);
                    // 取消消息发给对方
                    VoiceChatCancelMsg cancelMsgToTargetUser = new VoiceChatCancelMsg();
                    cancelMsgToTargetUser.setHisId(hisId);
                    cancelMsgToTargetUser.setUserId(userId);
                    cancelMsgToTargetUser.setTargetUserId(targetUserId);
                    cancelMsgToTargetUser.setHisCancelTime(curTime);
                    cancelMsgToTargetUser.setMsg("对方已取消");
                    imService.sendCustomImMsg(null, targetUserId, cancelMsgToTargetUser, null);
                }

                chatResponse.setSuccess(true);
                chatResponse.setMsg("停止成功");
                return chatResponse;
            } else if (ChatStatusEnum.Calling != statusEnum) {
                // 不是呼叫中
                log.info("语音速配hisId{},不支持停止", hisId);
                chatResponse.setSuccess(false);
                chatResponse.setMsg(statusEnum.getDesc() + "不支持停止");
                return chatResponse;
            }

            // 通话中 => 停止
            // 获取计费分钟数
            int hisMinutes = Convert.toInt(redisService.hashGet(RedisFinanceKeys.VOICE_CHAT_CHARGING, Convert.toStr(hisId)), 1);
            int result = voiceChatHisDAO.updateStatusWithMinutes(hisId, ChatStatusEnum.Closed.getValue(), ChatStatusEnum.Calling.getValue(), hisStopTime, hisStopTime, hisMinutes, reasonEnum.getValue());
            if (result == 0) {
                log.info("修改状态失败, hisId {}", hisId);
                chatResponse.setSuccess(false);
                chatResponse.setMsg("修改状态失败");
                return chatResponse;
            }

            // 更新语音状态
            userService.updateVoiceChatStatus(userId, VoiceChatStatus.IDLE);
            userService.updateVoiceChatStatus(targetUserId, VoiceChatStatus.IDLE);

            Integer hisCoinMinutes = hisMinutes - hisTimeMinutes;

            // 停止mq
            QueueVoiceChatStopMsg stopMsg = new QueueVoiceChatStopMsg();
            stopMsg.setHisId(hisId);
            stopMsg.setHisMinutes(hisMinutes);
            stopMsg.setHisStopTime(hisStopTime);
            stopMsg.setReasonEnum(reasonEnum);
            stopMsg.setAppVer(appVer);
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.VOICE_CHAT_STOP, stopMsg.serialize());

            VoiceChatStopMsg stopMsgToUser = new VoiceChatStopMsg();
            stopMsgToUser.setHisId(hisId);
            stopMsgToUser.setUserId(userId);
            stopMsgToUser.setTargetUserId(targetUserId);
            stopMsgToUser.setStopReason(reasonEnum.getValue());
            stopMsgToUser.setStopReasonName(reasonEnum.getDesc());
            stopMsgToUser.setHisMinutes(hisMinutes);
            stopMsgToUser.setHisCoinMinutes(hisCoinMinutes);
            stopMsgToUser.setHisPrice(hisPrice);
            stopMsgToUser.setHisAmount(hisCoinMinutes * hisPrice);
            stopMsgToUser.setHisTimeMinutes(hisTimeMinutes);
            stopMsgToUser.setHisAmountName(WalletTypeEnum.toEnum(hisPriceType).getName());
            stopMsgToUser.setHisStartTime(hisStartTime);
            stopMsgToUser.setHisStopTime(hisStopTime);
            stopMsgToUser.setMsg("已结束");
            imService.sendCustomImMsg(null, userId, stopMsgToUser, null);

            VoiceChatStopMsg stopMsgToTargetUser = new VoiceChatStopMsg();
            stopMsgToTargetUser.setHisId(hisId);
            stopMsgToTargetUser.setUserId(userId);
            stopMsgToTargetUser.setTargetUserId(targetUserId);
            stopMsgToTargetUser.setStopReason(reasonEnum.getValue());
            stopMsgToTargetUser.setStopReasonName(reasonEnum.getDesc());
            stopMsgToTargetUser.setHisMinutes(hisMinutes);
            stopMsgToTargetUser.setHisCoinMinutes(hisCoinMinutes);
            stopMsgToTargetUser.setHisPrice(hisPrice);
            stopMsgToTargetUser.setHisAmount(Convert.toLong(hisCoinMinutes * hisPrice * 0.7));
            stopMsgToTargetUser.setHisTimeMinutes(hisTimeMinutes);
            stopMsgToTargetUser.setHisAmountName(WalletTypeEnum.toEnum(hisPriceType).getName());
            stopMsgToTargetUser.setHisStartTime(hisStartTime);
            stopMsgToTargetUser.setHisStopTime(hisStopTime);
            stopMsgToTargetUser.setMsg("已结束");
            imService.sendCustomImMsg(null, targetUserId, stopMsgToTargetUser, null);

            chatResponse.setSuccess(true);
            chatResponse.setMsg("处理成功");
            return chatResponse;
        } catch (Exception e) {
            log.info("出现异常，{}", e);
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
        }
        chatResponse.setSuccess(false);
        chatResponse.setMsg("处理失败");
        return chatResponse;
    }

    /**
     * 获取免费时长
     *
     * @param userId
     * @return
     */
    @Override
    public boolean hasFreeTime(Long userId, Long curTime) {
        return voiceChatProvider.hasFreeTime(userId, curTime);
    }

    /**
     * 暂停语音速配
     *
     * @param hisId
     * @param times
     * @return
     */
    @Override
    public ChatResponse cancelVoiceChat(Long hisId, int times) {
        ChatResponse chatResponse = new ChatResponse();
        log.info("暂停语音速配,hisId{}", hisId);

        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);
        if (hisDO == null) {
            log.info("未找到记录{}", hisId);
            try {
                if (times < 5) {
                    Thread.sleep(500L);
                    return cancelVoiceChat(hisId, times++);
                }
            } catch (Exception e) {
                log.info("出现异常", e);
            }
            chatResponse.setSuccess(false);
            chatResponse.setMsg("未找到记录");
            return chatResponse;
        }

        ChatStatusEnum statusEnum = ChatStatusEnum.toEnum(hisDO.getHisStatus());
        if (statusEnum != ChatStatusEnum.Ringing) {
            chatResponse.setSuccess(false);
            chatResponse.setMsg("当前速配" + statusEnum.getDesc());
            return chatResponse;
        }

        Long userId = hisDO.getUserId();
        // 更新用户聊天状态
        userService.updateVoiceChatStatus(userId, VoiceChatStatus.IDLE);

        // 统计记录语音速配匹配时间
        long matchStartTime = NumberUtils.toLong(redisService.get(RedisChatKeys.getVoiceChatMatchStartTime(userId)), 0);
        if (matchStartTime > 0) {
            long matchDuration = System.currentTimeMillis() - matchStartTime;
            statPlatformService.addPlatformStat(hisDO.getChannelId(), StatPlatformTypeEnum.VOICE_CHAT_MATCH_DURATION, new BigDecimal("" + matchDuration));
        }

        long curTime = System.currentTimeMillis();

        VoiceChatHisDO hisUpdate = new VoiceChatHisDO();
        hisUpdate.setHisId(hisId);
        hisUpdate.setHisStatus(ChatStatusEnum.Canceled.getValue());
        hisUpdate.setHisCancelTime(curTime);
        hisUpdate.setHisUpdateTime(curTime);
        if (voiceChatHisDAO.updateByIdAndWhereStatus(hisUpdate, ChatStatusEnum.Ringing.getValue()) > 0) {
            log.info("取消消息发给自己，hisId{}", hisId);
            Long targetUserId = hisDO.getTargetUserId();
            // 取消消息发给自己
            VoiceChatCancelMsg cancelMsgToUser = new VoiceChatCancelMsg();
            cancelMsgToUser.setHisId(hisId);
            cancelMsgToUser.setUserId(userId);
            cancelMsgToUser.setTargetUserId(targetUserId);
            cancelMsgToUser.setHisCancelTime(curTime);
            cancelMsgToUser.setMsg("您已取消");
            imService.sendCustomImMsg(null, userId, cancelMsgToUser, null);

            // 如果匹配到用户
            if (targetUserId != null) {
                log.info("取消消息发给对方，hisId{},targetUserId{}", hisId, targetUserId);
                userService.updateVoiceChatStatus(targetUserId, VoiceChatStatus.IDLE);
                // 取消消息发给对方
                VoiceChatCancelMsg cancelMsgToTargetUser = new VoiceChatCancelMsg();
                cancelMsgToTargetUser.setHisId(hisId);
                cancelMsgToTargetUser.setUserId(userId);
                cancelMsgToTargetUser.setTargetUserId(targetUserId);
                cancelMsgToTargetUser.setHisCancelTime(curTime);
                cancelMsgToTargetUser.setMsg("对方已取消");
                imService.sendCustomImMsg(null, targetUserId, cancelMsgToTargetUser, null);
            }

            chatResponse.setSuccess(true);
            chatResponse.setMsg("处理成功");
            return chatResponse;
        }
        chatResponse.setSuccess(false);
        chatResponse.setMsg("系统繁忙，请稍后重试");
        return chatResponse;
    }

    /**
     * 是否有权限连接
     *
     * @param cn
     * @param userId
     * @return
     */
    @Override
    public boolean hasConnectPermission(String cn, Long userId) {
        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(Convert.toLong(cn));
        if (hisDO != null) {
            if (ChatStatusEnum.Calling != ChatStatusEnum.toEnum(hisDO.getHisStatus())) {
                // 非通话状态不可加入
                return false;
            }
            if (hisDO.getUserId().equals(userId)) {
                // 用户可以加入
                return true;
            }
            if (hisDO.getTargetUserId().equals(userId)) {
                // 对方也可以加入
                return true;
            }
        }
        return false;
    }

    /**
     * 获取语音速配token
     *
     * @param channelName
     * @param userId
     * @param isPub
     * @return
     */
    @Override
    public String getVoiceChatToken(String channelName, Long userId, boolean isPub, String appVer) {
        return providerBuilder.provider(appVer).buildToken(channelName, userId, isPub ? AgoraProvider.Role.Role_Publisher.initValue : AgoraProvider.Role.Role_Subscriber.initValue, 0);
    }

    /**
     * 获取语音速配列表
     *
     * @param pageIndex
     * @param pageSize
     * @param hisStatus
     * @param userId
     * @return
     */
    @Override
    public PageList<VoiceChatDTO> getVoiceChatList(Integer pageIndex, int pageSize, Integer hisStatus, Long userId) {
        PageList<VoiceChatDTO> page = new PageList<>(pageIndex, pageSize);
        page.setList(voiceChatHisDAO.selectVoiceChatList(page.getPageCursor(), pageSize, hisStatus, userId));
        page.setTotal(voiceChatHisDAO.countVoiceChatList(hisStatus, userId));
        return page;
    }

    /**
     * 判断记录是否存在
     *
     * @param hisId
     * @return
     */
    @Override
    public boolean isHisExist(Long hisId) {
        return voiceChatHisDAO.countById(hisId) > 0;
    }

    /**
     * 获取发起人
     *
     * @param hisId
     * @return
     */
    @Override
    public Long getUserId(Long hisId) {
        return voiceChatHisDAO.selectUserId(hisId);
    }

    /**
     * 获取记录id
     *
     * @param userId
     * @return
     */
    @Override
    public Long getHisIdByUserId(Long userId) {
        return voiceChatHisDAO.selectHisId(userId);
    }

    /**
     * 获取匹配成功的记录
     */
    @Override
    public List<UserMatchVO> getMatchSuccessList() {
        return voiceChatHisDAO.selectMatchSuccessList();
    }

    /**
     * 获取符合条件用户头像
     *
     * @param userSex
     * @param userId
     * @return
     */
    @Override
    public List<UserMatchResponse> getMatchUserList(Integer userSex, Long userId) {
        return voiceChatProvider.getMatchUserAvatarList(userSex, Arrays.asList(userId));
    }

    /**
     * 获取匹配到的陪陪
     *
     * @param userSex 0:未知; 1:男; 2:女
     * @return
     */
    @Override
    public List<UserPeiPeiVO> getMatchPeipiList(Integer userSex) {
        return voiceChatProvider.getMatchPeipeiList(userSex, getMaxShowPeipeiCount());
    }

    /**
     * 设置交友匹配到的陪陪最大展示数量
     *
     * @param count
     * @return
     */
    @Override
    public void setMaxShowPeipeiCount(int count) {
        redisService.set(RedisChatKeys.MAX_SHOW_PEIPEI_COUNT, count + "");
    }

    /**
     * 获得交友匹配到的陪陪最大展示数量
     *
     * @return
     */
    @Override
    public int getMaxShowPeipeiCount() {
        String value = redisService.get(RedisChatKeys.MAX_SHOW_PEIPEI_COUNT);
        int count = NumberUtils.toInt(value);
        return count < 12 ? 12 : count;
    }

    /**
     * 获取语音速配匹配条数
     *
     * @return
     */
    @Override
    public int getMatchCount() {
        return Convert.toInt(redisService.get(RedisChatKeys.VOICE_CHAT_MATCH_COUNT), 10);
    }

    /**
     * 设置语音速配匹配条数
     *
     * @param count
     * @return
     */
    @Override
    public void setMatchCount(int count) {
        redisService.set(RedisChatKeys.VOICE_CHAT_MATCH_COUNT, Convert.toStr(count));
    }

    /**
     * 获取当前轮次
     *
     * @param hisId
     * @return
     */
    @Override
    public int getCurRound(Long hisId) {
        return voiceChatProvider.getCurRound(hisId);
    }

    /**
     * 获取用户轮次
     *
     * @param hisId
     * @param userId
     * @return
     */
    @Override
    public int getUserRound(Long hisId, Long userId) {
        Map<Object, Object> roundUsersMap = redisService.hashGetAll(RedisChatKeys.getVoiceChatRoundUser(hisId));
        if (CollectionUtil.isNotEmpty(roundUsersMap)) {
            for (Map.Entry<Object, Object> entry : roundUsersMap.entrySet()) {
                String value = Convert.toStr(entry.getValue());
                if (StringUtils.isNotEmpty(value) && value.contains(Convert.toStr(userId))) {
                    return Convert.toInt(entry.getKey());
                }
            }
        }
        return -1;
    }

    /**
     * 获取授权信息
     *
     * @param roomId
     * @param userId
     * @return
     */
    @Override
    public RtcAuthInfo getAuthInfo(Long roomId, Long userId, String appVer) {
        return providerBuilder.provider(appVer).getAuthInfo(roomId, userId);
    }
}
