package com.starbuds.server.service.media.provider;

import cn.hutool.core.convert.Convert;
import cn.hutool.http.HttpRequest;
import cn.hutool.http.HttpResponse;
import cn.hutool.http.HttpUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.rtc.AgroaRole;
import com.starbuds.server.common.constant.rtc.RtcProviderEnum;
import com.starbuds.server.common.pojo.data.QiniuConfig;
import com.starbuds.server.common.pojo.rtc.RtcAuthInfo;
import com.starbuds.server.common.pojo.rtc.RtcProviderSignature;
import com.starbuds.server.common.service.data.FileService;
import com.starbuds.server.common.service.media.ProviderInterface;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.agora.AccessToken;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.apache.dubbo.rpc.protocol.rest.support.ContentType;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.util.Arrays;

/**
 * 声网服务
 *
 * @author Lay
 * @date 2020/3/22
 */
@Component
@Slf4j
public class AgoraProvider implements ProviderInterface {


    public enum Role {
        /**
         * DEPRECATED. Role_Attendee has the same privileges as Role_Publisher.
         */
        Role_Attendee(0),
        /**
         * RECOMMENDED. Use this role for a voice/video call or a live broadcast, if your scenario does not require authentication for [Hosting-in](https://docs.agora.io/en/Agora%20Platform/terms?platform=All%20Platforms#hosting-in).
         */
        Role_Publisher(1),
        /**
         * Only use this role if your scenario require authentication for [Hosting-in](https://docs.agora.io/en/Agora%20Platform/terms?platform=All%20Platforms#hosting-in).
         *
         * @note In order for this role to take effect, please contact our support team to enable authentication for Hosting-in for you. Otherwise, Role_Subscriber still has the same privileges as Role_Publisher.
         */
        Role_Subscriber(2),
        /**
         * DEPRECATED. Role_Attendee has the same privileges as Role_Publisher.
         */
        Role_Admin(101);

        public int initValue;

        Role(int initValue) {
            this.initValue = initValue;
        }

        public static Role toEnum(Integer value) {
            if (value != null) {
                for (Role obj : Role.values()) {
                    if (obj.initValue == value) {
                        return obj;
                    }
                }
            }

            return Role.Role_Subscriber;
        }
    }

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    FileService fileService;

    @Value("${agora.appId}")
    private String appId;

    @Value("${agora.appCertificate}")
    private String appCertificate;

    @Value("${agora.baseUrl}")
    private String baseUrl;

    @Value("${agora.base64Credentials}")
    private String base64Credentials;

    @Value("${agora.secretRTC}")
    private String secretRTC;

    @Value("${agora.secretCloudRecording}")
    private String secretCloudRecording;


    @Resource
    ProviderBuilder providerBuilder;

    @PostConstruct
    public void init() {
        providerBuilder.regist(RtcProviderEnum.AGORA, this);
    }

    @Override
    /**
     * 获取授权信息
     * @param roomId
     * @param userId
     * @return
     */
    public RtcAuthInfo getAuthInfo(Long roomId, Long userId) {
        Long userNo = userService.getUserNo(userId);
        if (userNo != null && userNo > 0) {
            String channelName = String.format("YL-%d", roomId);
            RtcAuthInfo rtcAuthInfo = new RtcAuthInfo();
            rtcAuthInfo.setChannelName(channelName);
            rtcAuthInfo.setToken(buildToken(channelName, String.valueOf(userNo), AgroaRole.Role_Admin.initValue, System.currentTimeMillis() + TimeConstants.DAY_MILLS));
            rtcAuthInfo.setUserNo(Convert.toInt(userNo));
            return rtcAuthInfo;
        }
        return null;
    }

    @Override
    public String voiceChatChannelName(Long roomId, Long userId) {
        return Convert.toStr(roomId);
    }

    /**
     * 创建令牌
     *
     * @param channelName
     * @param userId
     * @param role
     * @param expireTime
     * @return
     */
    @Override
    public String buildToken(String channelName, Long userId, int role, long expireTime) {
        Long userNo = userService.getUserNo(userId);
        String account = userNo == null || userNo == 0 ? "" : String.valueOf(userNo);
        return buildToken(channelName, account, role, expireTime);
    }

    private String buildToken(String channelName, String account, int role, long expireTime) {
        AccessToken builder = new AccessToken(appId, appCertificate, channelName, account);
        int privilegeTs = Convert.toInt(expireTime / 1000, 0);
        builder.addPrivilege(AccessToken.Privileges.kJoinChannel, privilegeTs);
        if (role == Role.Role_Publisher.initValue || role == Role.Role_Subscriber.initValue || role == Role.Role_Admin.initValue) {
            builder.addPrivilege(AccessToken.Privileges.kPublishAudioStream, privilegeTs);
            builder.addPrivilege(AccessToken.Privileges.kPublishVideoStream, privilegeTs);
            builder.addPrivilege(AccessToken.Privileges.kPublishDataStream, privilegeTs);
        }

        try {
            return builder.build();
        } catch (Exception e) {
            e.printStackTrace();
            return "";
        }
    }

    /**
     * 禁用单聊频道
     *
     * @param channelName
     * @return
     */
    @Override
    public boolean disableChannel(String channelName) {
        JSONObject jsonObject = new JSONObject();
        jsonObject.put("appid", appId);
        jsonObject.put("cname", channelName);
        jsonObject.put("time", 1440);
        jsonObject.put("privileges", Arrays.asList("join_channel"));
        HttpRequest request = HttpUtil.createPost(String.format("%s/dev/v1/kicking-rule/", baseUrl))
                .header("Authorization", String.format("Basic %s", base64Credentials))
                .body(jsonObject.toJSONString(), ContentType.APPLICATION_JSON_UTF_8);
        HttpResponse response = request.execute();
        log.info("封禁单聊频道 {}", response.body());
        if (response.isOk()) {
            JSONObject result = JSON.parseObject(response.body());
            if (result != null && result.getString("status").equals("success")) {
                return true;
            }
        }
        return false;
    }

    /**
     * 判断用户是否在频道中
     *
     * @param channelName
     * @param userId
     * @return
     */
    @Override
    public boolean isInChannel(String channelName, Long userId) {
        log.info("--AgoraProvider : isInChannel");
        Long userNo = userService.getUserNo(userId);
        if (userNo == null || userNo == 0) {
            return false;
        }
        HttpRequest request = HttpUtil.createGet(String.format("%s/dev/v1/channel/user/property/%s/%d/%s", baseUrl, appId, userNo, channelName))
                .header("Authorization", String.format("Basic %s", base64Credentials));
        HttpResponse response = request.execute();
        if (response.isOk()) {
            log.info("获取状态 {}", response.body());
            JSONObject result = JSON.parseObject(response.body());
            if (result != null && result.getBoolean("success")) {
                return result.getJSONObject("data").getBoolean("in_channel");
            }
        }
        return false;
    }

    /**
     * 开始录制
     *
     * @param channelName
     * @return
     */
    @Override
    public String startRecord(String channelName) {
        if (true) {
            //功能不用
            return null;

        }

        QiniuConfig qiniuConfig = fileService.getQiniuConfig();
        if (qiniuConfig == null) {
            return null;
        }
        Long adminId = 10000L;
        String resourceId = getRecordAcquire(channelName, adminId);
        if (StringUtils.isNotBlank(resourceId)) {
            JSONObject params = new JSONObject();
            params.put("uid", String.valueOf(adminId));
            params.put("cname", channelName);

            JSONObject recordingConfig = new JSONObject();
            recordingConfig.put("channelType", 1);
            recordingConfig.put("streamTypes", 2);
            recordingConfig.put("audioProfile", 1);
            recordingConfig.put("videoStreamType", 1);
            recordingConfig.put("maxIdleTime", 10);

            JSONObject storageConfig = new JSONObject();
            storageConfig.put("vendor", 0);
            storageConfig.put("region", 1);
            storageConfig.put("bucket", qiniuConfig.getBucketVod());
            storageConfig.put("accessKey", qiniuConfig.getAccessKey());
            storageConfig.put("secretKey", qiniuConfig.getSecretKey());
            storageConfig.put("fileNamePrefix", Arrays.asList("agora"));

            JSONObject clientRequest = new JSONObject();
            clientRequest.put("token", buildToken(channelName, String.valueOf(adminId), Role.Role_Admin.initValue, System.currentTimeMillis() + TimeConstants.DAY_MILLS));
            clientRequest.put("recordingConfig", recordingConfig);
            clientRequest.put("storageConfig", storageConfig);

            params.put("clientRequest", clientRequest);

            String uri = String.format("%s/v1/apps/%s/cloud_recording/resourceid/%s/mode/%s/start", baseUrl, appId, resourceId, "mix");
            log.info("请求开始录制参数 {} {}", uri, params.toJSONString());
            HttpRequest request = HttpUtil.createPost(uri)
                    .header("Authorization", String.format("Basic %s", base64Credentials))
                    .body(params.toJSONString(), ContentType.APPLICATION_JSON_UTF_8);
            HttpResponse response = request.execute();
            log.info("请求开始录制 {} {}", response.getStatus(), response.body());
            if (response.isOk()) {
                return response.body();
            }
        }
        return null;
    }

    /**
     * 停止录制
     *
     * @param channelName
     * @return
     */
    @Override
    public String stopRecord(String channelName, String resourceId, String sid) {
        if (true) {
            //功能不用
            return null;

        }
        if (StringUtils.isNotBlank(resourceId)) {
            JSONObject params = new JSONObject();
            params.put("cname", channelName);
            params.put("uid", String.valueOf(10000L));
            params.put("clientRequest", new JSONObject());
            String uri = String.format("%s/v1/apps/%s/cloud_recording/resourceid/%s/sid/%s/mode/%s/stop", baseUrl, appId, resourceId, sid, "mix");
            log.info("请求结束录制 {} 参数 {}", uri, params.toJSONString());
            HttpRequest request = HttpUtil.createPost(uri)
                    .header("Authorization", String.format("Basic %s", base64Credentials))
                    .body(params.toJSONString(), ContentType.APPLICATION_JSON_UTF_8);
            HttpResponse response = request.execute();
            log.info("请求结束录制 {} {}", response.getStatus(), response.body());
            if (response.isOk()) {
                JSONObject body = JSON.parseObject(response.body());
                return body.getJSONObject("serverResponse").getString("fileList");
            }
        }
        return null;
    }

    private String getRecordAcquire(String channelName, Long userNo) {
        JSONObject jsonObject = new JSONObject();
        jsonObject.put("cname", channelName);
        jsonObject.put("uid", String.valueOf(userNo));
        JSONObject clientRequest = new JSONObject();
        clientRequest.put("resourceExpiredHour", 24);
        jsonObject.put("clientRequest", clientRequest);
        HttpRequest request = HttpUtil.createPost(String.format("%s/v1/apps/%s/cloud_recording/acquire", baseUrl, appId))
                .header("Authorization", String.format("Basic %s", base64Credentials))
                .body(jsonObject.toJSONString(), ContentType.APPLICATION_JSON_UTF_8);
        HttpResponse response = request.execute();
        log.info("请求 resourceId {}", response.body());
        if (response.isOk()) {
            JSONObject result = JSON.parseObject(response.body());
            return result.getString("resourceId");
        }
        return null;
    }

    @Override
    public boolean verifyNotifySign(RtcProviderSignature signature) {
        return hmacSha1(signature.getNotify(), secretRTC).equalsIgnoreCase(signature.getSignature()) || hmacSha1(signature.getNotify(), secretCloudRecording).equalsIgnoreCase(signature.getSignature());
    }

    /**
     * 将加密后的字节数组转换成字符串
     *
     * @param bytes
     * @return
     */
    private static String bytesToHex(byte[] bytes) {
        StringBuffer sb = new StringBuffer();

        for (int i = 0; i < bytes.length; i++) {
            String hex = Integer.toHexString(bytes[i] & 0xFF);

            if (hex.length() < 2) {
                sb.append(0);
            }

            sb.append(hex);
        }

        return sb.toString();
    }

    /**
     * HMAC/SHA1 加密，返回加密后的字符串
     *
     * @param message
     * @param secret
     * @return
     */
    private static String hmacSha1(String message, String secret) {
        try {
            SecretKeySpec signingKey = new SecretKeySpec(secret.getBytes(
                    "utf-8"), "HmacSHA1");
            Mac mac = Mac.getInstance("HmacSHA1");
            mac.init(signingKey);

            byte[] rawHmac = mac.doFinal(message.getBytes("utf-8"));

            return bytesToHex(rawHmac);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

}
