package com.starbuds.server.service.media.provider;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import cn.hutool.crypto.SecureUtil;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.cache.RedisLiveKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.media.BattleRoleEnum;
import com.starbuds.server.common.constant.media.BattleStatusEnum;
import com.starbuds.server.common.constant.media.BattlingStatus;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.core.Tuple;
import com.starbuds.server.common.pojo.db.LiveBattleAnchorDO;
import com.starbuds.server.common.pojo.db.LiveBattleDO;
import com.starbuds.server.common.pojo.db.LiveBattleGiverDO;
import com.starbuds.server.common.pojo.message.im.live.LiveBattleSnapshotMsg;
import com.starbuds.server.common.pojo.message.im.live.LiveBattleStatusMsg;
import com.starbuds.server.common.pojo.message.im.live.LiveBattleTimeoutMsg;
import com.starbuds.server.common.pojo.message.im.live.LiveBattleValueMsg;
import com.starbuds.server.common.pojo.mq.QueueBattleStageDelayMsg;
import com.starbuds.server.common.pojo.mq.QueueBattleWaitTimeoutMsg;
import com.starbuds.server.common.pojo.user.UserCardVO;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.AnchorDAO;
import com.starbuds.server.service.media.dao.LiveBattleAnchorDAO;
import com.starbuds.server.service.media.dao.LiveBattleDAO;
import com.starbuds.server.service.media.dao.LiveBattleGiverDAO;
import com.starbuds.server.service.media.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.util.*;

/**
 * PK 服务
 *
 * @author Lay
 * @date 2020/4/20
 */
@Component
@Slf4j
public class BattleProvider {

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Resource
    LiveBattleDAO liveBattleDAO;

    @Resource
    AnchorDAO anchorDAO;

    @Resource
    LiveBattleAnchorDAO liveBattleAnchorDAO;

    @Resource
    LiveBattleGiverDAO liveBattleGiverDAO;

    @Resource
    RocketProducer rocketProducer;

    /**
     * PK 时间（秒）
     */
    public static final Long BATTLE_TIME = 420L;

    /**
     * PK 惩罚时间（秒）
     */
    public static final Long PUNISH_TIME = 180L;

    /**
     * PK 邀请超时
     *
     * @param battleId
     * @param userId
     * @param destUserId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public boolean battleInviteTimeout(String battleId, Long userId, Long destUserId) {
        try {
            if (liveBattleDAO.updateBattleStatus(battleId, BattleStatusEnum.Timeout.getValue(), BattleStatusEnum.Ringing.getValue()) > 0) {
                // 恢复主播状态
                anchorDAO.updateBattling(userId, BoolIntVal.False, BoolIntVal.True);
                anchorDAO.updateBattling(destUserId, BoolIntVal.False, BoolIntVal.True);

                LiveBattleTimeoutMsg timeoutMsg = new LiveBattleTimeoutMsg();
                timeoutMsg.setBattleId(battleId);

                // 发送超时消息
                imService.sendLiveRoomMsg(ImTargetType.USER, null, userId, timeoutMsg);
                imService.sendLiveRoomMsg(ImTargetType.USER, null, destUserId, timeoutMsg);
                return true;
            }
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
        }
        return false;
    }

    @Transactional(rollbackFor = Exception.class)
    public boolean startBattle(LiveBattleDO liveBattle) {
        // 尝试更新状态
        if (liveBattleDAO.updateBattleStatus(liveBattle.getBattleId(), BattleStatusEnum.Battling.getValue(), BattleStatusEnum.Ringing.getValue()) > 0) {
            // 创建主播记录
            LiveBattleAnchorDO anchorRed = new LiveBattleAnchorDO();
            anchorRed.setBattleId(liveBattle.getBattleId());
            anchorRed.setBattleRole(BattleRoleEnum.Red.getValue());
            anchorRed.setUserId(liveBattle.getUserId());

            LiveBattleAnchorDO anchorBlue = new LiveBattleAnchorDO();
            anchorBlue.setBattleId(liveBattle.getBattleId());
            anchorBlue.setBattleRole(BattleRoleEnum.Blue.getValue());
            anchorBlue.setUserId(liveBattle.getDestUserId());

            liveBattleAnchorDAO.batchInsert(Arrays.asList(anchorRed, anchorBlue));

            // 等待加入
            redisService.hashPut(RedisLiveKeys.LIVE_BATTLE_STATUS, liveBattle.getBattleId(), String.valueOf(BattlingStatus.Wait));

            // 等待加入超时检查
            QueueBattleWaitTimeoutMsg timeoutMsg = new QueueBattleWaitTimeoutMsg();
            timeoutMsg.setBattleId(liveBattle.getBattleId());
            rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.BATTLE_WAIT_TIMEOUT, timeoutMsg.serialize(), System.currentTimeMillis() + TimeConstants.SECOND_15_MILLS);

            return true;
        }
        return false;
    }

    /**
     * 加入 PK 房间
     *
     * @param battleId
     * @param userId
     * @return
     */
    public boolean joinBattleRoom(String battleId, Long userId) {
        LiveBattleDO liveBattle = liveBattleDAO.selectById(battleId);
        if (liveBattle == null) {
            return false;
        }

        // 判断状态
        Object status = redisService.hashGet(RedisLiveKeys.LIVE_BATTLE_STATUS, battleId);
        if (status != null) {
            Long curTime = System.currentTimeMillis();
            int battlingStatus = Convert.toInt(status);
            if (BattlingStatus.Wait == battlingStatus) {
                // 把主播丢进去
                if (redisService.setIfAbsent(RedisLiveKeys.getLiveBattleAnchor(userId), battleId)) {
                    // 判断是否到齐了
                    if (redisService.hasKey(RedisLiveKeys.getLiveBattleAnchor(liveBattle.getUserId())) && redisService.hasKey(RedisLiveKeys.getLiveBattleAnchor(liveBattle.getDestUserId()))) {
                        // 更新开战时间
                        liveBattleDAO.updateStartTime(battleId, curTime);

                        // 开战
                        redisService.hashPut(RedisLiveKeys.getLiveBattleValue(battleId), String.valueOf(liveBattle.getUserId()), "0");
                        redisService.hashPut(RedisLiveKeys.getLiveBattleValue(battleId), String.valueOf(liveBattle.getDestUserId()), "0");
                        redisService.hashPut(RedisLiveKeys.LIVE_BATTLE_STATUS, battleId, String.valueOf(BattlingStatus.Battling));

                        UserCardVO userCardRed = userService.getCard(liveBattle.getUserId());
                        UserCardVO userCardBlue = userService.getCard(liveBattle.getDestUserId());

                        // 发送消息到直播间
                        LiveBattleStatusMsg statusMsgRed = new LiveBattleStatusMsg();
                        statusMsgRed.setBattleId(battleId);
                        statusMsgRed.setBattlingStatus(BattlingStatus.Battling);
                        statusMsgRed.setBattleTime(BATTLE_TIME);
                        statusMsgRed.setPunishTime(PUNISH_TIME);
                        LiveBattleStatusMsg statusMsgBlue = new LiveBattleStatusMsg();
                        statusMsgBlue.setBattleId(battleId);
                        statusMsgBlue.setBattlingStatus(BattlingStatus.Battling);
                        statusMsgBlue.setBattleTime(BATTLE_TIME);
                        statusMsgBlue.setPunishTime(PUNISH_TIME);

                        // 红方消息
                        statusMsgRed.setBattleRole(BattleRoleEnum.Red.getValue());
                        statusMsgRed.setDestUserId(liveBattle.getDestUserId());
                        statusMsgRed.setDestUserName(userCardBlue.getUserName());
                        statusMsgRed.setDestUserAvatar(userCardBlue.getUserAvatar());

                        // 蓝方消息
                        statusMsgBlue.setBattleRole(BattleRoleEnum.Blue.getValue());
                        statusMsgBlue.setDestUserId(liveBattle.getUserId());
                        statusMsgBlue.setDestUserName(userCardRed.getUserName());
                        statusMsgBlue.setDestUserAvatar(userCardRed.getUserAvatar());

                        // 发送双方直播间
                        imService.sendLiveRoomMsg(ImTargetType.GROUP, null, liveBattle.getUserId(), statusMsgRed);
                        imService.sendLiveRoomMsg(ImTargetType.GROUP, null, liveBattle.getDestUserId(), statusMsgBlue);

                        // 惩罚阶段计时
                        QueueBattleStageDelayMsg stageMsg = new QueueBattleStageDelayMsg();
                        stageMsg.setBattleId(battleId);
                        stageMsg.setStage(BattlingStatus.Punish);
                        rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.BATTLE_STAGE_DELAY, stageMsg.serialize(), curTime + BATTLE_TIME * 1000L);

                        return true;
                    }
                }
            }
        }

        return false;
    }

    /**
     * 获取 PK 状态
     *
     * @param battleId
     * @return
     */
    public Integer getBattlingStatus(String battleId) {
        Object status = redisService.hashGet(RedisLiveKeys.LIVE_BATTLE_STATUS, battleId);
        if (status != null) {
            return Convert.toInt(status);
        }
        return null;
    }

    /**
     * 获取 PK 值
     *
     * @param battleId
     * @param userId
     * @return
     */
    public Long getBattleValue(String battleId, Long userId) {
        return Optional.ofNullable(redisService.hashGet(RedisLiveKeys.getLiveBattleValue(battleId), String.valueOf(userId))).map(Convert::toLong).orElse(0L);
    }

    /**
     * PK 结算
     *
     * @param battleId
     */
    public void battleSettle(String battleId) {
        Integer battlingStatus = getBattlingStatus(battleId);
        if (battlingStatus == null) {
            return;
        }

        if (BattlingStatus.Battling == getBattlingStatus(battleId)) {
            // 修改 pk 状态
            redisService.hashPut(RedisLiveKeys.LIVE_BATTLE_STATUS, battleId, String.valueOf(BattlingStatus.Punish));

            LiveBattleDO liveBattle = liveBattleDAO.selectById(battleId);

            Long userIdRed = liveBattle.getUserId();
            Long userIdBlue = liveBattle.getDestUserId();

            // 获取双方值
            Long redValue = getBattleValue(battleId, userIdRed);
            Long blueValue = getBattleValue(battleId, userIdBlue);

            // 判断输赢
            int result = 0;
            if (redValue > blueValue) {
                result = BattleRoleEnum.Red.getValue();
            } else if (blueValue > redValue) {
                result = BattleRoleEnum.Blue.getValue();
            }

            battlePersis(battleId, userIdRed, userIdBlue, result, redValue, blueValue);
        }
    }

    public void battlePersis(String battleId, Long userIdRed, Long userIdBlue, int result, Long redValue, Long blueValue) {
        Long curTime = System.currentTimeMillis();

        // 发送结果
        LiveBattleStatusMsg statusMsg = new LiveBattleStatusMsg();
        statusMsg.setBattleId(battleId);
        statusMsg.setBattlingStatus(BattlingStatus.Punish);
        statusMsg.setBattleTime(BATTLE_TIME);
        statusMsg.setPunishTime(PUNISH_TIME);
        statusMsg.setBattleResult(result);
        statusMsg.setTimestamp(curTime);
        imService.sendLiveRoomMsg(ImTargetType.GROUP, null, userIdRed, statusMsg);
        imService.sendLiveRoomMsg(ImTargetType.GROUP, null, userIdBlue, statusMsg);

        // 惩罚阶段延迟
        QueueBattleStageDelayMsg stageDelayMsg = new QueueBattleStageDelayMsg();
        stageDelayMsg.setBattleId(battleId);
        stageDelayMsg.setStage(BattlingStatus.Over);
        rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.BATTLE_STAGE_DELAY, stageDelayMsg.serialize(), curTime + PUNISH_TIME * 1000L);

        // 更新主播信息
        if (result == 0) {
            // 平局
            anchorDAO.incrBattleDrawTimes(Arrays.asList(userIdRed, userIdBlue), 1);
            // 更新战斗结果
            liveBattleAnchorDAO.updateResult(battleId, userIdRed, redValue, 0);
            liveBattleAnchorDAO.updateResult(battleId, userIdBlue, blueValue, 0);
        } else if (result == BattleRoleEnum.Red.getValue()) {
            // 红方胜
            anchorDAO.incrBattleWinTimes(userIdRed, 1);
            anchorDAO.incrBattleLossTimes(userIdBlue, 1);
            // 更新战斗结果
            liveBattleAnchorDAO.updateResult(battleId, userIdRed, redValue, 1);
            liveBattleAnchorDAO.updateResult(battleId, userIdBlue, blueValue, -1);
        } else {
            // 蓝方胜
            anchorDAO.incrBattleWinTimes(userIdBlue, 1);
            anchorDAO.incrBattleLossTimes(userIdRed, 1);
            // 更新战斗结果
            liveBattleAnchorDAO.updateResult(battleId, userIdRed, redValue, -1);
            liveBattleAnchorDAO.updateResult(battleId, userIdBlue, blueValue, 1);
        }

        // pk次数
        anchorDAO.incrBattleTimes(userIdRed, 1);
        anchorDAO.incrBattleTimes(userIdBlue, 1);

        // 更新战斗结果
        liveBattleDAO.updateBattleResult(battleId, result, curTime);

        // 持久化用户贡献
        saveGiver(battleId, userIdRed);
        saveGiver(battleId, userIdBlue);
    }

    /**
     * 持久化贡献者
     *
     * @param battleId
     * @param userId
     */
    public void saveGiver(String battleId, Long userId) {
        Set<Tuple> tupleSetRed = redisService.zReverseRangeWithScores(RedisLiveKeys.getLiveBattleGiver(battleId, userId), 0, -1);
        if (tupleSetRed != null && tupleSetRed.size() > 0) {
            List<Tuple> list = new ArrayList<>(tupleSetRed);
            list.sort(Comparator.comparingDouble(Tuple::getScore).reversed());
            List<LiveBattleGiverDO> giverList = new ArrayList<>();
            int rankNo = 1;
            for (Tuple tuple : list) {
                LiveBattleGiverDO liveBattleGiver = new LiveBattleGiverDO();
                liveBattleGiver.setBattleId(battleId);
                liveBattleGiver.setAnchorUserId(userId);
                liveBattleGiver.setUserId(Convert.toLong(tuple.getElement()));
                liveBattleGiver.setGiveScore(Convert.toLong(tuple.getScore()));
                liveBattleGiver.setRankNo(rankNo);
                giverList.add(liveBattleGiver);

                rankNo++;
            }

            // 保存
            liveBattleGiverDAO.batchInsert(giverList);
        }
    }

    /**
     * PK 结束
     *
     * @param battleId
     */
    public boolean battleOver(String battleId) {
        LiveBattleDO liveBattle = liveBattleDAO.selectById(battleId);
        if (liveBattle != null) {
            Integer battlingStatus = getBattlingStatus(battleId);
            if (battlingStatus == null) {
                return false;
            }
            Integer battleStatus = BattleStatusEnum.Over.getValue();
            if (BattlingStatus.Wait == battlingStatus) {
                battleStatus = BattleStatusEnum.Canceled.getValue();
            }

            // 修改状态
            liveBattleDAO.updateBattleOver(battleId, battleStatus, System.currentTimeMillis());

            // 恢复主播状态
            anchorDAO.updateBattling(liveBattle.getUserId(), BoolIntVal.False, BoolIntVal.True);
            anchorDAO.updateBattling(liveBattle.getDestUserId(), BoolIntVal.False, BoolIntVal.True);

            // 发送结束消息
            LiveBattleStatusMsg statusMsg = new LiveBattleStatusMsg();
            statusMsg.setBattleId(battleId);
            statusMsg.setBattlingStatus(BattlingStatus.Over);
            imService.sendLiveRoomMsg(ImTargetType.GROUP, null, liveBattle.getUserId(), statusMsg);
            imService.sendLiveRoomMsg(ImTargetType.GROUP, null, liveBattle.getDestUserId(), statusMsg);

            // 一些清理工作
            redisService.hashDel(RedisLiveKeys.LIVE_BATTLE_STATUS, battleId);
            redisService.delKeys(Arrays.asList(
                    RedisLiveKeys.getLiveBattleProfile(battleId),
                    RedisLiveKeys.getLiveBattleValue(battleId),
                    RedisLiveKeys.getLiveBattleAnchor(liveBattle.getUserId()),
                    RedisLiveKeys.getLiveBattleAnchor(liveBattle.getDestUserId()),
                    RedisLiveKeys.getLiveBattleGiver(battleId, liveBattle.getUserId()),
                    RedisLiveKeys.getLiveBattleGiver(battleId, liveBattle.getDestUserId())
            ));
            return true;
        }
        return false;
    }

    /**
     * 认输
     *
     * @param battleId
     * @param userId
     * @return
     */
    public boolean surrender(String battleId, Long userId) {
        if (BattlingStatus.Battling == getBattlingStatus(battleId)) {
            // 修改 pk 状态
            redisService.hashPut(RedisLiveKeys.LIVE_BATTLE_STATUS, battleId, String.valueOf(BattlingStatus.Punish));

            LiveBattleDO liveBattle = liveBattleDAO.selectById(battleId);

            Long userIdRed = liveBattle.getUserId();
            Long userIdBlue = liveBattle.getDestUserId();

            // 获取双方值
            Long redValue = getBattleValue(battleId, userIdRed);
            Long blueValue = getBattleValue(battleId, userIdBlue);

            // 判断输赢
            int result = liveBattle.getUserId().equals(userId) ? BattleRoleEnum.Blue.getValue() : BattleRoleEnum.Red.getValue();

            battlePersis(battleId, userIdRed, userIdBlue, result, redValue, blueValue);
            return true;
        }
        return false;
    }

    /**
     * 增加 PK 值
     *
     * @param anchorUserId
     * @param userId
     * @param value
     */
    @Async
    public void addValueAsync(Long anchorUserId, Long userId, Long value) {
        String battleId = Optional.ofNullable(redisService.get(RedisLiveKeys.getLiveBattleAnchor(anchorUserId))).map(String::toString).orElse("");
        if (StringUtils.isNotBlank(battleId)) {
            // 判断状态
            if (BattlingStatus.Battling == Optional.ofNullable(getBattlingStatus(battleId)).orElse(BattlingStatus.Over)) {
                // 增加 PK 条
                Long battleValue = redisService.hashIncreaseLong(RedisLiveKeys.getLiveBattleValue(battleId), String.valueOf(anchorUserId), value);

                // 增加用户值
                redisService.zIncrementScore(RedisLiveKeys.getLiveBattleGiver(battleId, anchorUserId), String.valueOf(userId), value);

                // 缓存用户
                cacheGiver(battleId, userId);

                // 发送值
                LiveBattleValueMsg valueMsg = new LiveBattleValueMsg();
                valueMsg.setUserId(anchorUserId);
                valueMsg.setBattleValue(battleValue);
                valueMsg.setTopGivers(getTopGivers(battleId, anchorUserId, 4));
                imService.sendBattleGroupMsg(battleId, valueMsg);
            }
        }
    }

    /**
     * 缓存贡献者
     *
     * @param battleId
     * @param userId
     */
    public void cacheGiver(String battleId, Long userId) {
        if (!redisService.hashExists(RedisLiveKeys.getLiveBattleProfile(battleId), String.valueOf(userId))) {
            List<UserCardVO> list = userService.getNameAndAvatarByIds(Arrays.asList(userId));
            if (CollectionUtil.isNotEmpty(list)) {
                redisService.hashPut(RedisLiveKeys.getLiveBattleProfile(battleId), String.valueOf(userId), JSON.toJSONString(list.get(0)));
            }
        }
    }

    /**
     * 获取贡献者资料
     *
     * @param battleId
     * @param userId
     * @return
     */
    public UserCardVO getGiver(String battleId, Long userId) {
        Object user = redisService.hashGet(RedisLiveKeys.getLiveBattleProfile(battleId), String.valueOf(userId));
        if (user != null) {
            return JSON.parseObject(user.toString(), UserCardVO.class);
        }
        return null;
    }

    /**
     * 根据用户 id 结束 PK
     *
     * @param userId
     * @return
     */
    public boolean battleOverByUserId(Long userId) {
        LiveBattleDO liveBattle = liveBattleDAO.selectLatestByUserId(userId, BattleStatusEnum.Battling.getValue());
        if (liveBattle != null) {
            // 结束掉
            if (BattlingStatus.Battling == Optional.ofNullable(getBattlingStatus(liveBattle.getBattleId())).orElse(BattlingStatus.Punish)) {
                // 认输
                return surrender(liveBattle.getBattleId(), userId);
            }
            return battleOver(liveBattle.getBattleId());
        }
        return false;
    }

    /**
     * 进入 PK 场景
     *
     * @param anchorUserId
     * @param userId
     */
    @Async
    public void goBattleScene(Long anchorUserId, Long userId) {
        LiveBattleDO liveBattle = liveBattleDAO.selectLatestByUserId(anchorUserId, BattleStatusEnum.Battling.getValue());
        log.info("进入 PK 场景判断 {} {}", anchorUserId, userId);
        if (liveBattle != null) {
            log.info("进入 PK 场景成功 {}", liveBattle.getBattleId());
            BattleRoleEnum battleRole;
            Long destUserId;
            Long redValue;
            Long blueValue;

            if (liveBattle.getUserId().equals(anchorUserId)) {
                battleRole = BattleRoleEnum.Red;
                destUserId = liveBattle.getDestUserId();
                redValue = getBattleValue(liveBattle.getBattleId(), liveBattle.getUserId());
                blueValue = getBattleValue(liveBattle.getBattleId(), destUserId);
            } else {
                battleRole = BattleRoleEnum.Blue;
                destUserId = liveBattle.getUserId();
                blueValue = getBattleValue(liveBattle.getBattleId(), liveBattle.getDestUserId());
                redValue = getBattleValue(liveBattle.getBattleId(), destUserId);
            }

            UserCardVO userCard = userService.getCard(destUserId);
            Integer battlingStatus = getBattlingStatus(liveBattle.getBattleId());
            LiveBattleSnapshotMsg snapshotMsg = new LiveBattleSnapshotMsg();
            snapshotMsg.setDestUserId(destUserId);
            snapshotMsg.setDestUserName(userCard.getUserName());
            snapshotMsg.setDestUserAvatar(userCard.getUserAvatar());
            snapshotMsg.setBattlingStatus(battlingStatus);
            snapshotMsg.setBattleRole(battleRole.getValue());
            snapshotMsg.setRedValue(redValue);
            snapshotMsg.setBlueValue(blueValue);
            snapshotMsg.setBattleResult(liveBattle.getBattleResult());
            snapshotMsg.setRedGivers(getTopGivers(liveBattle.getBattleId(), liveBattle.getUserId(), 4));
            snapshotMsg.setBlueGivers(getTopGivers(liveBattle.getBattleId(), liveBattle.getDestUserId(), 4));

            Long timeRemain = null;
            if (BattlingStatus.Battling == battlingStatus) {
                timeRemain = liveBattle.getBattleStartTime() + BATTLE_TIME * 1000L - System.currentTimeMillis();
            } else if (BattlingStatus.Punish == battlingStatus) {
                timeRemain = liveBattle.getBattleSettleTime() + PUNISH_TIME * 1000L - System.currentTimeMillis();
            }
            snapshotMsg.setTimeRemain(timeRemain);
            imService.sendLiveRoomMsg(ImTargetType.USER, null, userId, snapshotMsg);
        }
    }

    public List<UserCardVO> getTopGivers(String battleId, Long anchorUserId, int num) {
        Set<Tuple> giverSet = redisService.zReverseRangeWithScores(RedisLiveKeys.getLiveBattleGiver(battleId, anchorUserId), 0, num);
        if (CollectionUtil.isNotEmpty(giverSet)) {
            List<Tuple> giverList = new ArrayList<>(giverSet);
            giverList.sort(Comparator.comparingDouble(Tuple::getScore).reversed());
            List<UserCardVO> topGivers = new ArrayList<>();
            for (Tuple tuple : giverList) {
                Long userId = Convert.toLong(tuple.getElement());
                UserCardVO userCard = getGiver(battleId, userId);
                if (userCard != null) {
                    topGivers.add(userCard);
                }
            }
            return topGivers;
        }
        return Collections.emptyList();
    }

    public static String getBattleId(Long userId, Long destUserId, Long time) {
        return SecureUtil.md5(String.format("%d:%d:%d", userId, destUserId, time));
    }
}
