package com.starbuds.server.service.media.provider;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import cn.hutool.core.util.RandomUtil;
import cn.hutool.crypto.SecureUtil;
import com.starbuds.server.common.constant.cache.RedisCacheKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.finance.ChatTimeWalletType;
import com.starbuds.server.common.constant.media.LiveStatusEnum;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.db.AnchorDO;
import com.starbuds.server.common.pojo.db.SingleChatMatchHisDO;
import com.starbuds.server.common.pojo.message.im.chat.SingleChatMatchMsg;
import com.starbuds.server.common.pojo.mq.QueueSingleChatMatchDelayMsg;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.finance.VipService;
import com.starbuds.server.common.service.finance.WalletService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.AnchorDAO;
import com.starbuds.server.service.media.dao.SingleChatFreeHisDAO;
import com.starbuds.server.service.media.dao.SingleChatMatchHisDAO;
import com.starbuds.server.service.media.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

/**
 * 聊天匹配服务
 *
 * @author Lay
 * @date 2020/4/7
 */
@Component
@Slf4j
public class ChatMatchProvider {

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    VipService vipService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    WalletService walletService;

    @Resource
    RocketProducer rocketProducer;

    @Resource
    AnchorDAO anchorDAO;

    @Resource
    SingleChatMatchHisDAO singleChatMatchHisDAO;

    @Resource
    SingleChatFreeHisDAO singleChatFreeHisDAO;

    /**
     * 将用户加入匹配池
     *
     * @param userId
     * @return
     */
    public boolean addMatchPoolUser(Long userId) {
        // 前置条件检查
        if (!checkUserPoolCondition(userId)) {
            return false;
        }

        // 加入匹配池
        return redisService.sAdd(RedisCacheKeys.SINGLE_CHAT_MATCH_POOL_USER, String.valueOf(userId)) == 1;
    }

    /**
     * 将用户从匹配池移除
     *
     * @param userId
     * @return
     */
    public boolean removeMatchPoolUser(Long userId) {
        return redisService.sRemove(RedisCacheKeys.SINGLE_CHAT_MATCH_POOL_USER, String.valueOf(userId)) == 1;
    }

    /**
     * 从匹配池随机取出用户
     *
     * @return
     */
    public Long popPoolUser() {
        String userId = redisService.sPop(RedisCacheKeys.SINGLE_CHAT_MATCH_POOL_USER);
        if (StringUtils.isNotBlank(userId)) {
            return Convert.toLong(userId);
        }
        return null;
    }

    /**
     * 将主播加入匹配池
     *
     * @param userId
     * @return
     */
    public boolean addMatchPoolAnchor(Long userId) {
        // 前置条件检查
        if (!checkAnchorPoolCondition(userId)) {
            return false;
        }

        // 加入匹配池
        return redisService.sAdd(RedisCacheKeys.SINGLE_CHAT_MATCH_POOL_ANCHOR, String.valueOf(userId)) == 1;
    }

    /**
     * 将主播从匹配池移除
     *
     * @param userId
     * @return
     */
    public boolean removeMatchPoolAnchor(Long userId) {
        return redisService.sRemove(RedisCacheKeys.SINGLE_CHAT_MATCH_POOL_ANCHOR, String.valueOf(userId)) == 1;
    }

    /**
     * 从匹配池随机取出主播
     *
     * @return
     */
    public Long popPoolAnchor() {
        String userId = redisService.sPop(RedisCacheKeys.SINGLE_CHAT_MATCH_POOL_ANCHOR);
        if (StringUtils.isNotBlank(userId)) {
            return Convert.toLong(userId);
        }
        return null;
    }

    /**
     * 获取一个机器人主播
     *
     * @return
     */
    public Long getRobotAnchor() {
        List<Long> robotAnchors = anchorDAO.selectSingleChatRobotAnchorIds();
        if (CollectionUtil.isNotEmpty(robotAnchors)) {
            int randomIndex = RandomUtil.randomInt(0, robotAnchors.size());
            return robotAnchors.get(randomIndex);
        }
        return null;
    }

    /**
     * 检查用户匹配池条件
     *
     * @param userId
     * @return
     */
    private boolean checkUserPoolCondition(Long userId) {
        // 加入条件判断 非主播 非 VIP 、未充值用户
        // 判断是否主播
        if (anchorDAO.countById(userId) > 0) {
            return false;
        }

        // 是否充值过
        boolean isRecharged = userService.hasRecharged(userId);
        if (isRecharged) {
            return false;
        }
        // 是否 VIP
        boolean isVip = vipService.isVip(userId);
        if (isVip) {
            return false;
        }

        // 领取过时间且用完的
        if (singleChatFreeHisDAO.countById(userId) > 0 && walletService.getChatTimeWalletMinutes(userId, ChatTimeWalletType.SINGLE_CHAT) == 0) {
            return false;
        }

        // 是否在直播间
        if (redisService.hashExists(RedisCacheKeys.LIVE_USER_ROOM, String.valueOf(userId))) {
            return false;
        }

        return true;
    }

    /**
     * 检查主播匹配池条件
     *
     * @param userId
     * @return
     */
    private boolean checkAnchorPoolCondition(Long userId) {
        // 非主播 pass
        AnchorDO anchorDO = anchorDAO.selectById(userId);
        if (anchorDO == null) {
            return false;
        }

        // 没有 1v1 权限 pass
        if (BoolIntVal.True != Optional.ofNullable(anchorDO.getSingleChatEnable()).orElse(BoolIntVal.False)) {
            return false;
        }

        // 不能匹配 pass
        if (BoolIntVal.True != Optional.ofNullable(anchorDO.getSingleChatMatchEnable()).orElse(BoolIntVal.False)) {
            return false;
        }

        // 直播中 pass
        if (LiveStatusEnum.Living == LiveStatusEnum.toEnum(anchorDO.getLiveStatus())) {
            return false;
        }

        return true;
    }

    /**
     * 自动匹配
     *
     * @return
     */
    public boolean matchAuto() {
        // 从用户池取一个用户
        Long userId = popPoolUser();
        return matchUser(userId);
    }

    /**
     * 延迟匹配
     *
     * @param userId
     * @param delay
     */
    public void matchDelay(Long userId, Long delay) {
        addMatchPoolUser(userId);
        QueueSingleChatMatchDelayMsg delayMsg = new QueueSingleChatMatchDelayMsg();
        delayMsg.setUserId(userId);
        rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.SINGLE_CHAT_MATCH_DELAY, delayMsg.serialize(), System.currentTimeMillis() + delay);
    }

    /**
     * 为指定用户匹配
     *
     * @param userId
     * @return
     */
    public boolean matchUser(Long userId) {
        log.debug("为指定用户匹配 {}", userId);
        if (removeMatchPoolUser(userId)) {
            // 从主播池取一个主播
            boolean isRobot = false;
            Long anchorUserId = popPoolAnchor();
            log.debug("从主播池取一个主播 {}", anchorUserId);
            if (anchorUserId == null) {
                // 取一个机器人主播
                anchorUserId = getRobotAnchor();
                log.debug("取一个机器人主播 {}", anchorUserId);
                isRobot = true;
            }

            // 机器人主播也没得
            if (anchorUserId == null) {
                log.debug("机器人主播也没得");
                return false;
            }

            // 匹配成功
            return startMatch(userId, anchorUserId, isRobot);
        }
        return false;
    }

    /**
     * 开始匹配
     *
     * @param userId
     * @param anchorUserId
     * @param isRobot
     * @return
     */
    public boolean startMatch(Long userId, Long anchorUserId, boolean isRobot) {
        log.debug("开始匹配 userId:{} anchorUserId:{} isRobot:{}", userId, anchorUserId, isRobot);
        Long curTime = System.currentTimeMillis();

        // 再次检查条件
        if (!checkUserPoolCondition(userId)) {
            return false;
        }

        // 锁用户 30s
        log.debug("锁用户 30s {}", userId);
        if (redisService.setIfAbsent(RedisCacheKeys.getSingleChatMatchLock(userId), String.valueOf(anchorUserId), TimeConstants.SECOND_30_MILLS, TimeUnit.MILLISECONDS)) {
            String matchId;
            // 机器人匹配模式
            if (isRobot) {
                log.debug("机器人匹配模式 {}", userId);
                matchId = getMatchId(userId, anchorUserId, curTime);

                // 给用户发送视频聊天消息
                log.debug("给用户发送视频聊天消息 {}", userId);
                SingleChatMatchMsg matchMsg = new SingleChatMatchMsg();
                matchMsg.setAnchorUserId(anchorUserId);
                matchMsg.setUserId(userId);
                matchMsg.setTimestamp(curTime);
                matchMsg.setShowVip(vipService.isVip(userId) ? BoolIntVal.False : BoolIntVal.True);
                imService.sendCustomImMsg(null, userId, matchMsg, null);

                // 延迟给用户匹配
                matchDelay(userId, TimeConstants.MIN_MILLS * 5L);
            } else {
                log.debug("真人匹配模式 {}", userId);
                // 再次检查主播
                if (!checkAnchorPoolCondition(anchorUserId)) {
                    // 重新匹配
                    log.debug("重新匹配 {}", userId);
                    matchUser(userId);
                    return false;
                }
                // 真实主播匹配，锁定主播 30s
                log.debug("真实主播匹配，锁定主播 30s {}", userId);
                if (redisService.setIfAbsent(RedisCacheKeys.getSingleChatMatchLock(anchorUserId), String.valueOf(userId), TimeConstants.SECOND_30_MILLS, TimeUnit.MILLISECONDS)) {
                    matchId = getMatchId(userId, anchorUserId, curTime);

                    // 给主播发送消息
                    log.debug("给主播发送消息");
                    SingleChatMatchMsg matchMsg = new SingleChatMatchMsg();
                    matchMsg.setMatchId(matchId);
                    matchMsg.setAnchorUserId(anchorUserId);
                    matchMsg.setUserId(userId);
                    matchMsg.setTimestamp(curTime);
                    imService.sendCustomImMsg(null, anchorUserId, matchMsg, null);
                } else {
                    // 主播锁定失败
                    matchUser(userId);
                    addMatchPoolAnchor(anchorUserId);
                    return false;
                }
            }

            // 创建匹配记录
            if (matchId != null) {
                SingleChatMatchHisDO singleChatMatchHis = new SingleChatMatchHisDO();
                singleChatMatchHis.setMatchId(matchId);
                singleChatMatchHis.setUserId(userId);
                singleChatMatchHis.setAnchorUserId(anchorUserId);
                singleChatMatchHis.setIsRobot(isRobot ? BoolIntVal.True : BoolIntVal.False);
                singleChatMatchHis.setMatchTime(System.currentTimeMillis());
                singleChatMatchHisDAO.insert(singleChatMatchHis);
            }

            return true;
        }
        return false;
    }

    /**
     * 获取匹配 id
     *
     * @param userId
     * @param anchorUserId
     * @param time
     * @return
     */
    private static String getMatchId(Long userId, Long anchorUserId, Long time) {
        return SecureUtil.md5(String.format("%d:%d:%d", anchorUserId, userId, time));
    }
}
