package com.starbuds.server.service.media.provider;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.cache.RedisLiveKeys;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.pojo.core.Tuple;
import com.starbuds.server.common.pojo.media.LiveHeat;
import com.starbuds.server.common.pojo.message.im.live.LiveHeatChangeMsg;
import com.starbuds.server.common.pojo.mq.QueueLiveHeatExpireMsg;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.*;

/**
 * 直播热度算法
 *
 * @author Lay
 * @date 2019/12/25
 */
@Component
@Slf4j
public class LiveHeatProvider {

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Resource
    RocketProducer rocketProducer;

    /**
     * 初始化热度值
     *
     * @param userId
     */
    @Async
    public void initLiveHeatAsync(Long userId) {
        long heat = 0L;

        // 重置主播热度值
        removeLiveHeat(userId);

        // 根据主播等级换算热度值
        int charmLevel = userService.getCharmLevelByUserId(userId);
        int wealthLevel = userService.getWealthLevelByUserId(userId);
        // 主播魅力等级加成
        heat += charmLevel * 200 * 0.75;
        // 主播财富等级加成
        heat += wealthLevel * 50 * 0.25;

        // 改变热度值
        changeLiveHeat(userId, heat, 0, true);

        // 设置热度标记
        setHeatKey(userId);
    }

    /**
     * 移除热度
     *
     * @param userId
     */
    public boolean removeLiveHeat(Long userId) {
        try {
            // 移除热度榜
            redisService.zRemove(RedisLiveKeys.LIVE_HEAT_SET, String.valueOf(userId));

            // 移除最高热度
            redisService.hashDel(RedisLiveKeys.LIVE_HEAT_MAX, String.valueOf(userId));

            // 移除 key
            removeHeatKey(userId);
        } catch (Exception e) {
            log.error("热度移除失败", e);
        }
        return false;
    }

    /**
     * 更新热度值
     *
     * @param userId
     * @param heatValue
     * @param duration
     * @return
     */
    public Long changeLiveHeat(Long userId, long heatValue, long duration) {
        return changeLiveHeat(userId, heatValue, duration, false);
    }

    /**
     * 更新热度值
     *
     * @param userId
     * @param heatValue
     * @param duration
     * @param ignoreExists
     * @return
     */
    public Long changeLiveHeat(Long userId, long heatValue, long duration, boolean ignoreExists) {
        try {
            if (!ignoreExists) {
                if (redisService.zScore(RedisLiveKeys.LIVE_HEAT_SET, String.valueOf(userId)) == null) {
                    return 0L;
                }
            }

            // 写入热度值
            Double curValue = redisService.zIncrementScore(RedisLiveKeys.LIVE_HEAT_SET, String.valueOf(userId), Convert.toDouble(heatValue));

            if (duration > 0) {
                // 通过延时消息队列实现热度值减持
                rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.LIVE_HEAT_EXPIRE, QueueLiveHeatExpireMsg.builder().userId(userId).heatValue(heatValue).heatKey(getHeatKey(userId)).build().serialize(), System.currentTimeMillis() + duration);
            }

            Long heatValueNew = Convert.toLong(curValue);

            // 判断本次直播最高热度
            if (heatValueNew > Optional.ofNullable(redisService.hashGet(RedisLiveKeys.LIVE_HEAT_MAX, String.valueOf(userId))).map(t -> Convert.toLong(t)).orElse(0L)) {
                redisService.hashPut(RedisLiveKeys.LIVE_HEAT_MAX, String.valueOf(userId), String.valueOf(heatValueNew));
            }

            // 发送到直播间
            LiveHeatChangeMsg liveHeatChangeMsg = new LiveHeatChangeMsg();
            liveHeatChangeMsg.setHeatValue(heatValueNew);
            imService.sendLiveRoomMsg(ImTargetType.GROUP, null, userId, liveHeatChangeMsg);

            return heatValueNew;
        } catch (Exception e) {
            log.error("热度值更新失败", e);
        }
        return null;
    }

    /**
     * 获取热门列表
     *
     * @param num
     * @return
     */
    public List<LiveHeat> getTopN(int start, int num) {
        Set<Tuple> tuples = redisService.zReverseRangeWithScores(RedisLiveKeys.LIVE_HEAT_SET, start, start + num - 1);
        if (CollectionUtil.isNotEmpty(tuples)) {
            List<LiveHeat> liveHeats = new ArrayList<>();
            for (Tuple tuple : tuples) {
                LiveHeat liveHeat = new LiveHeat();
                liveHeat.setUserId(Convert.toLong(tuple.getElement()));
                liveHeat.setHeatValue(Convert.toLong(tuple.getScore()));
                liveHeats.add(liveHeat);
            }
            liveHeats.sort(Comparator.comparingLong(LiveHeat::getHeatValue).reversed());
            return liveHeats;
        }
        return Collections.EMPTY_LIST;
    }

    /**
     * 获取用户的热度
     *
     * @param userId 用户id
     * @return score
     */
    public Long getLiveHeat(Long userId) {
        return Optional.ofNullable(redisService.zScore(RedisLiveKeys.LIVE_HEAT_SET, String.valueOf(userId))).orElse(0d).longValue();
    }

    /**
     * 获取峰值热度
     *
     * @param userId
     * @return
     */
    public Long getLiveHeatMax(Long userId) {
        return Optional.ofNullable(redisService.hashGet(RedisLiveKeys.LIVE_HEAT_MAX, String.valueOf(userId))).map(t -> Convert.toLong(t)).orElse(0L);
    }

    private void setHeatKey(Long userId) {
        redisService.hashPut(RedisLiveKeys.LIVE_HEAT_KEY, String.valueOf(userId), String.valueOf(System.currentTimeMillis()));
    }

    public String getHeatKey(Long userId) {
        Object obj = redisService.hashGet(RedisLiveKeys.LIVE_HEAT_KEY, String.valueOf(userId));
        if (obj != null) {
            return obj.toString();
        }
        return null;
    }

    private void removeHeatKey(Long userId) {
        redisService.hashDel(RedisLiveKeys.LIVE_HEAT_KEY, String.valueOf(userId));
    }
}
