package com.starbuds.server.service.media.provider;

import cn.hutool.crypto.SecureUtil;
import com.qiniu.pili.Client;
import com.qiniu.pili.Hub;
import com.qiniu.rtc.RtcRoomManager;
import com.qiniu.util.Auth;
import com.starbuds.server.common.pojo.media.LiveStreamPullUrl;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

/**
 * 七牛直播服务
 *
 * @author Lay
 * @date 2019/12/23
 */
@Component
@Slf4j
public class QiniuLiveProvider {
    private String accessKey;
    private String secretKey;
    private String rtcAppId;
    private Client client;
    private Hub hub;

    @Value("${qiniu.hubname}")
    private String hubName;

    @Value("${qiniu.livekey}")
    private String liveKey;

    @Value("${qiniu.domain.publish}")
    private String qiniuDomainPublish;

    @Value("${qiniu.domain.snapshot}")
    private String qiniuDomainSnapshot;

    @Value("${qiniu.domain.rtmp}")
    private String qiniuDomainRtmp;

    @Value("${qiniu.domain.hls}")
    private String qiniuDomainHls;

    @Value("${qiniu.domain.hdl}")
    private String qiniuDomainHdl;

    @Value("${qiniu.domain.vod}")
    private String qiniuDomainVod;

    @Value("${qiniu.liveSpace}")
    private String qiniuLiveSpace;

    public QiniuLiveProvider(@Value("${qiniu.accessKey}") String accessKey,
                             @Value("${qiniu.secretKey}") String secretKey,
                             @Value("${qiniu.rtcAppId}") String rtcAppId) {
        this.accessKey = accessKey;
        this.secretKey = secretKey;
        this.rtcAppId = rtcAppId;
        client = new Client(accessKey, secretKey);
    }

    public Hub getHub() {
        if (hub == null) {
            hub = client.newHub(hubName);
        }
        return hub;
    }

    public String getSign(String url, int expireSeconds) {
        String path = url.substring(url.indexOf(hubName) - 1);
        long time = System.currentTimeMillis() / 1000;
        time = time + expireSeconds;
        String toSign = String.format("%s%s%d", liveKey, path, time);
        String sign = SecureUtil.md5(toSign).toLowerCase();
        return String.format("%s?sign=%s&t=%d", url, sign, time);
    }

    public String getPushUrl(Long userId, Long liveHisId, Long streamHisId, int expireSeconds) {
        try {
            String pushUrl = client.RTMPPublishURL(qiniuDomainPublish, hubName, String.valueOf(userId), expireSeconds);
            return String.format("%s&hid=%d&sid=%d", pushUrl, liveHisId, streamHisId);
        } catch (Exception e) {
            log.error("推流地址创建失败", e);
        }
        return null;
    }

    public LiveStreamPullUrl getPullUrl(Long userId, int expireSeconds) {
        String rtmp = getSign(client.RTMPPlayURL(qiniuDomainRtmp, hubName, String.valueOf(userId)), expireSeconds);
        String hls = getSign(client.HLSPlayURL(qiniuDomainHls, hubName, String.valueOf(userId)), expireSeconds);
        String hdl = getSign(client.HDLPlayURL(qiniuDomainHdl, hubName, String.valueOf(userId)), expireSeconds);

        return LiveStreamPullUrl.builder().rtmp(rtmp).hls(hls).hdl(hdl).build();
    }

    public String saveStream(Long userId, Long streamStartTime, Long streamStopTime) {
        try {
            return String.format("http://%s/%s", qiniuDomainVod, getHub().get(String.valueOf(userId)).save(streamStartTime, streamStopTime));
        } catch (Exception e) {
            log.error("回放保存失败", e);
        }
        return null;
    }

    public boolean disableStream(Long userId, long disabledTill) {
        try {
            getHub().get(String.valueOf(userId)).disable(disabledTill);
            return true;
        } catch (Exception e) {
            log.error("禁播流失败", e);
        }
        return false;
    }

    public String getUserRoomToken(String roomName, Long userId, long expireTime) {
        return getRoomToken(roomName, userId, expireTime, "user");
    }

    public String getRoomToken(String roomName, Long userId, long expireTime, String role) {
        Auth auth = Auth.create(accessKey, secretKey);
        RtcRoomManager rmManager = new RtcRoomManager(auth);
        try {
            return rmManager.getRoomToken(rtcAppId, roomName, String.valueOf(userId), (System.currentTimeMillis() + expireTime) / 1000, role);
        } catch (Exception e) {
            log.error("RoomToken 生成失败", e);
        }
        return null;
    }

    public String getRtcAppId() {
        return rtcAppId;
    }

    public String getQiniuLiveSpace() {
        return qiniuLiveSpace;
    }
}
