package com.starbuds.server.service.media.provider;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.cache.RedisCacheKeys;
import com.starbuds.server.common.pojo.core.Tuple;
import com.starbuds.server.common.pojo.media.LiveRankItem;
import com.starbuds.server.common.pojo.user.UserCardVO;
import com.starbuds.server.common.service.core.CacheService;
import com.starbuds.server.common.service.user.UserService;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;

import java.text.SimpleDateFormat;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * 排行服务
 *
 * @author Lay
 * @date 2020/1/10
 */
@Component
@Slf4j
public class RankProvider {

    @Reference(version = "1.0.0")
    CacheService cacheService;

    @Reference(version = "1.0.0")
    UserService userService;

    static {
        StringBuilder sb = new StringBuilder();
        // 加日榜
        sb.append("redis.call('ZINCRBY', KEYS[1], ARGV[2], ARGV[1]);");
        // 周榜
        sb.append("redis.call('ZINCRBY', KEYS[2], ARGV[2], ARGV[1]);");
        // 月榜
        sb.append("redis.call('ZINCRBY', KEYS[3], ARGV[2], ARGV[1]);");
        LUA_INCR = sb.toString();
    }

    static {
        StringBuilder sb = new StringBuilder();
        sb.append("local t = redis.call('TIME');");
        sb.append("return t[1]");
        LUA_TIME = sb.toString();
    }

    /**
     * 中央时间戳
     */
    public static final String LUA_TIME;

    /**
     * 加榜
     */
    public static final String LUA_INCR;

    /**
     * 获取 Redis 时间戳
     *
     * @return
     */
    private long getRedisTime() {
        return cacheService.luaLong(LUA_TIME, new ArrayList<>(), new ArrayList<>());
    }

    public String[] getRankTimes() {
        long redisTime = getRedisTime();
        String[] times = new String[3];

        // 日
        Date date = new Date(redisTime * 1000);
        SimpleDateFormat sdfDay = new SimpleDateFormat("yyyyMMdd");
        times[0] = sdfDay.format(date);

        // 基础周为 2020年1月6日第一周
        long baseTime = 1578240000;
        long secDiff = redisTime - baseTime;
        double dayDiff = secDiff / 3600 / 24;
        int week = (int) (dayDiff / 7 + 1);
        times[1] = String.valueOf(week);

        // 月
        SimpleDateFormat sdfMonth = new SimpleDateFormat("yyyyMM");
        times[2] = sdfMonth.format(date);

        return times;
    }

    public List<LiveRankItem> getLiveRankItems(String key, long start, long end) {
        List<Tuple> rankData = new ArrayList<>(cacheService.zReverseRangeWithScores(key, start, end));

        if (CollectionUtil.isNotEmpty(rankData)) {
            List<LiveRankItem> liveRankItems = new ArrayList<>();

            // 需要加载的用户
            Set<Long> userIds = rankData.stream().map(t -> Convert.toLong(t.getElement())).collect(Collectors.toSet());

            // 拉取用户数据
            List<UserCardVO> userCardVOList = userService.getNameAndAvatarByIds(new ArrayList<>(userIds));
            Map<Long, UserCardVO> userMap = userCardVOList.stream().collect(Collectors.toMap(UserCardVO::getUserId, Function.identity()));
            int index = 0;

            rankData.sort((o1, o2) -> {
                if (o1.getScore().equals(o2.getScore())) {
                    return Convert.toLong(o1.getElement()) > Convert.toLong(o2.getElement()) ? 1 : -1;
                }
                return o1.getScore() > o2.getScore() ? -1 : 1;
            });

            UserCardVO defaultUser = new UserCardVO();
            defaultUser.setUserName("已注销");
            for (Tuple tuple : rankData) {
                index++;
                Long userId = Convert.toLong(tuple.getElement());
                LiveRankItem liveRankItem = new LiveRankItem();
                liveRankItem.setRankNo(index);
                liveRankItem.setUserId(userId);
                liveRankItem.setRankNo(index);
                liveRankItem.setAmount(Convert.toLong(tuple.getScore()));

                UserCardVO userCardVO = userMap.getOrDefault(liveRankItem.getUserId(), defaultUser);
                if (userMap.containsKey(liveRankItem.getUserId())) {
                    liveRankItem.setUserName(userCardVO.getUserName());
                    liveRankItem.setUserAvatar(userCardVO.getUserAvatar());
                    liveRankItem.setWealthLevel(userCardVO.getWealthLevel());
                }

                liveRankItems.add(liveRankItem);
            }

            return liveRankItems;
        }

        return Collections.emptyList();
    }

    public void addLiveRankAmount(Long anchorUserId, Long userId, Long amount) {
        if (amount < 1L) {
            return;
        }

        String[] times = getRankTimes();

        // 加榜
        List<String> keys = new ArrayList<>();
        keys.add(RedisCacheKeys.getLiveRankDayKey(anchorUserId, times[0]));
        keys.add(RedisCacheKeys.getLiveRankWeekKey(anchorUserId, times[1]));
        keys.add(RedisCacheKeys.getLiveRankMonthKey(anchorUserId, times[2]));

        // 参数
        List<String> args = new ArrayList<>();
        args.add(String.valueOf(userId));
        args.add(String.valueOf(amount));

        // 加榜
        cacheService.luaLong(RankProvider.LUA_INCR, keys, args);
    }
}
