package com.starbuds.server.service.media.provider;

import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.cache.RedisCacheKeys;
import com.starbuds.server.common.constant.cache.RedisFinanceKeys;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.media.ChatStatusEnum;
import com.starbuds.server.common.constant.media.SingleChatStopReasonEnum;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.constant.user.SingleChatStatus;
import com.starbuds.server.common.pojo.db.SingleChatAnchorMinutesHisDO;
import com.starbuds.server.common.pojo.db.SingleChatHisDO;
import com.starbuds.server.common.pojo.message.im.chat.SingleChatStopMsg;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.service.media.dao.AnchorDAO;
import com.starbuds.server.service.media.dao.SingleChatAnchorMinutesHisDAO;
import com.starbuds.server.service.media.dao.SingleChatHisDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import javax.annotation.Resource;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * 备注
 *
 * @author Lay
 * @date 2020/3/23
 */
@Component
@Slf4j
public class SingleChatProvider {

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Resource
    ProviderBuilder providerBuilder;


    @Resource
    ChatMatchProvider chatMatchProvider;

    @Resource
    SingleChatHisDAO singleChatHisDAO;

    @Resource
    AnchorDAO anchorDAO;

    @Resource
    SingleChatAnchorMinutesHisDAO singleChatAnchorMinutesHisDAO;

    /**
     * 停止聊天
     *
     * @param hisId
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public SingleChatHisDO stopChat(String hisId, SingleChatStopReasonEnum singleChatStopReason, String appVer) {
        try {
            SingleChatHisDO singleChatHis = singleChatHisDAO.selectById(hisId);
            if (singleChatHis == null) {
                return null;
            }

            // 不是通话中直接返回 ok
            if (ChatStatusEnum.Calling != ChatStatusEnum.toEnum(singleChatHis.getHisStatus())) {
                return singleChatHis;
            }

            // 获取计费分钟数
            int minutes = Convert.toInt(redisService.hashGet(RedisFinanceKeys.SINGLE_CHAT_CHARGING, hisId), 0);
            // 获取时间钱包分钟数
            int timeMinutes = Convert.toInt(redisService.hashGet(RedisFinanceKeys.SINGLE_CHAT_CHARGING_TIME, hisId), 0);
            int coinMinutes = minutes - timeMinutes;

            // 修改状态
            Long curTime = System.currentTimeMillis();
            int updated = singleChatHisDAO.updateStatusWithMinutes(hisId, ChatStatusEnum.Closed.getValue(), ChatStatusEnum.Calling.getValue(), curTime, minutes, coinMinutes, timeMinutes);
            if (updated == 0) {
                singleChatHis = singleChatHisDAO.selectById(hisId);
                return singleChatHis;
            }

            singleChatHis.setHisStatus(ChatStatusEnum.Closed.getValue());
            singleChatHis.setHisMinutes(minutes);
            singleChatHis.setHisCoinMinutes(coinMinutes);
            singleChatHis.setHisTimeMinutes(timeMinutes);
            singleChatHis.setHisUpdateTime(curTime);

            // 中止频道
            providerBuilder.provider(appVer).disableChannel(hisId);

            // 发送停止消息
            SingleChatStopMsg singleChatStopMsg = new SingleChatStopMsg();
            singleChatStopMsg.setHisId(hisId);
            singleChatStopMsg.setReasonCode(singleChatStopReason.getValue());
            singleChatStopMsg.setRemarks(singleChatStopReason.getDesc());
            singleChatStopMsg.setHisMinutes(minutes);
            singleChatStopMsg.setHisAmount(singleChatHis.getHisPrice() * coinMinutes);
            singleChatStopMsg.setHisAmountName(WalletTypeEnum.toEnum(singleChatHis.getHisPriceType()).getName());
            singleChatStopMsg.setHisCoinMinutes(coinMinutes);
            singleChatStopMsg.setHisTimeMinutes(timeMinutes);
            imService.sendChatGroupMsg(ImTargetType.GROUP, null, null, hisId, singleChatStopMsg);

            // 修改主播状态
            anchorDAO.updateSingleChatStatus(singleChatHis.getTargetUserId(), SingleChatStatus.IDLE, SingleChatStatus.BUSY);

            // 增加通话分钟数
            anchorDAO.increaseSingleChatMinutes(singleChatHis.getTargetUserId(), minutes);

            if (timeMinutes > 0) {
                // 增加主播的接听分钟数
                SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
                Integer date = Convert.toInt(sdf.format(new Date()));
                SingleChatAnchorMinutesHisDO singleChatAnchorMinutesHis = new SingleChatAnchorMinutesHisDO();
                singleChatAnchorMinutesHis.setHisDate(date);
                singleChatAnchorMinutesHis.setUserId(singleChatHis.getTargetUserId());
                singleChatAnchorMinutesHis.setMinutes(timeMinutes);
                singleChatAnchorMinutesHis.setUpdateTime(System.currentTimeMillis());
                singleChatAnchorMinutesHisDAO.insertOrUpdate(singleChatAnchorMinutesHis);
            }

            // 移除统计
            redisService.hashDel(RedisFinanceKeys.SINGLE_CHAT_CHARGING, hisId);
            redisService.hashDel(RedisFinanceKeys.SINGLE_CHAT_CHARGING_TIME, hisId);

            // 停止录制
            String hisReplay = providerBuilder.provider(appVer).stopRecord(singleChatHis.getHisId(), singleChatHis.getHisResourceId(), singleChatHis.getHisReplayId());
            if (StringUtils.isNotBlank(hisReplay)) {
                singleChatHisDAO.updateReplay(singleChatHis.getHisId(), hisReplay);
            }

            // 主播加回池子
            chatMatchProvider.addMatchPoolAnchor(singleChatHis.getTargetUserId());

            // 解除锁定
            redisService.delKey(RedisCacheKeys.getSingleChatMatchLock(singleChatHis.getUserId()));
            redisService.delKey(RedisCacheKeys.getSingleChatMatchLock(singleChatHis.getTargetUserId()));

            return singleChatHis;
        } catch (Exception e) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            log.error("单聊停止失败", e);
        }
        return null;
    }
}
