package com.starbuds.server.service.media.provider;

import cn.hutool.core.convert.Convert;
import cn.hutool.core.date.DateUtil;
import com.alibaba.fastjson.JSON;
import com.starbuds.server.common.constant.cache.RedisCacheKeys;
import com.starbuds.server.common.constant.cache.RedisChatKeys;
import com.starbuds.server.common.constant.common.BoolIntVal;
import com.starbuds.server.common.constant.data.MatchModeEnum;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.constant.user.VoiceChatStatus;
import com.starbuds.server.common.pojo.db.VoiceChatHisDO;
import com.starbuds.server.common.pojo.user.UserMatchResponse;
import com.starbuds.server.common.pojo.user.UserPeiPeiVO;
import com.starbuds.server.common.pojo.user.UserPeipeiDO;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.rtc.RtcRoomService;
import com.starbuds.server.common.service.rtc.RtcSeatService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.FamilySkillDAO;
import com.starbuds.server.service.media.dao.VoiceChatHisDAO;
import lombok.extern.slf4j.Slf4j;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * @Author jiangxiaowei
 * @DateTime 2020/11/16 10:25
 * @Description 语音速配
 */
@Slf4j
@Component
public class VoiceChatProvider {

    @Resource
    FamilySkillDAO familySkillDAO;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Reference(version = "1.0.0")
    RtcSeatService rtcSeatService;

    @Reference(version = "1.0.0")
    private RtcRoomService rtcRoomService;

    @Reference(version = "1.0.0")
    private StatPlatformService statPlatformService;

    @Resource
    private VoiceChatHisDAO voiceChatHisDAO;

    /**
     * 判断用户指定日期是否有免费时长
     *
     * @param userId
     * @param matchStartTime
     * @return
     */
    public boolean hasFreeTime(Long userId, Long matchStartTime) {
        return redisService.hashGet(RedisChatKeys.getVoiceChatFreeTimeKey(userId), DateUtil.format(new Date(matchStartTime), "yyyyMMdd")) == null;
    }

    /**
     * 使用免费时长
     *
     * @param userId
     * @param curTime
     */
    public void useFreeTime(Long userId, Long curTime) {
        redisService.hashIncreaseLong(RedisChatKeys.getVoiceChatFreeTimeKey(userId), DateUtil.format(new Date(curTime), "yyyyMMdd"), Convert.toLong(BoolIntVal.True));
    }

    /**
     * 用户是否空闲
     *
     * @param userId
     * @return
     */
    public boolean isUserIdle(Long userId) {
        return userService.getVoiceChatStatus(userId) == VoiceChatStatus.IDLE;
    }

    /**
     * 匹配用户
     *
     * @param hisId
     * @param userSex
     * @param matchShouldEndTime
     * @param matchNum
     * @return
     */
    public List<Long> matchUser(Long hisId, Integer userSex, Long matchShouldEndTime, int matchMode, int matchNum) {
        MatchModeEnum matchModeEnum = MatchModeEnum.toEnum(matchMode);
        List<Long> matchUserIds = new ArrayList<>();
        List<Long> excludedUserIds = getExcludedUserIds(hisId);
        switch (matchModeEnum) {
            case RANDOM:
                // 随机匹配
                List<Long> userIds = matchUserForRandom(userSex, excludedUserIds, matchNum);
                // 过滤麦位上用户
                List<Long> userIdList = new ArrayList<>();
                for (Long userId : userIds) {
                    if (rtcSeatService.isOnSeat(userId)) {
                        log.info("用户在麦上，跳过userId {}", userId);
                        addExcludedUserIds(hisId, userId);
                        continue;
                    }
                    userIdList.add(userId);
                }
                if (CollectionUtils.isEmpty(userIdList)) {
                    log.info("随机匹配，未匹配到 userSex {},excludedUserIds {}", userSex, excludedUserIds);
                    return Collections.EMPTY_LIST;
                }
                Random random = new Random();
                int n = random.nextInt(userIdList.size());
                matchUserIds.add(userIdList.get(n));
                break;
            case CONNECT_RATE:
                // 接通率匹配
                List<Long> userByConnectRate = matchUserByConnectRate(userSex, excludedUserIds, matchNum);
                if (CollectionUtils.isEmpty(userByConnectRate)) {
                    log.info("按照接通率匹配, 未匹配到用户");
                } else {
                    for (Long userId : userByConnectRate) {
                        // 过滤麦位上的用户
                        if (rtcSeatService.isOnSeat(userId)) {
                            log.info("用户在麦上，跳过userId {}", userId);
                            addExcludedUserIds(hisId, userId);
                        } else {
                            matchUserIds.add(userId);
                        }
                    }
                }
                break;
            case ACTIVE:
                // 接通率匹配
                List<Long> userByActive = matchUserByActive(userSex, excludedUserIds, matchNum);
                if (CollectionUtils.isEmpty(userByActive)) {
                    log.info("按照接通率匹配, 未匹配到用户 userSex {} , excludedUserIds {} ,matchNum {}", userSex, excludedUserIds, matchNum);
                } else {
                    for (Long userId : userByActive) {
                        // 过滤麦位上的用户
                        if (rtcSeatService.isOnSeat(userId)) {
                            log.info("用户在麦上，跳过userId {}", userId);
                            addExcludedUserIds(hisId, userId);
                        } else {
                            matchUserIds.add(userId);
                        }
                    }
                }
                break;
            default:
                log.error("入参错误，不支持此模式");
                break;
        }

        // 有剩余时间
        if (System.currentTimeMillis() < matchShouldEndTime) {
            // 未匹配到 => 继续匹配
            if (CollectionUtils.isEmpty(matchUserIds)) {
                try {
                    Thread.sleep(500L);
                    return matchUser(hisId, userSex, matchShouldEndTime, matchMode, matchNum);
                } catch (Exception e) {
                    log.info("出现错误", e);
                }
            } else {
                VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);
                if (hisDO != null) {
                    // 统计语音速配推送陪陪数
                    statPlatformService.addPlatformStat(hisDO.getChannelId(), StatPlatformTypeEnum.VOICE_CHAT_MATCH_SEND_PP_COUNT, new BigDecimal("" + matchUserIds.size()));
                }
                // 匹配到 => 返回结果
                return matchUserIds;
            }
        }

        return Collections.EMPTY_LIST;
    }

    /**
     * 随机匹配用户
     *
     * @param userSex
     * @param excludedUserIds
     * @return
     */
    private List<Long> matchUserForRandom(Integer userSex, List<Long> excludedUserIds, int matchNum) {
        log.info("随机匹配用户");
        return familySkillDAO.selectUserIds(userSex, excludedUserIds, matchNum);
    }

    /**
     * 根据接通率匹配
     *
     * @param userSex
     * @param excludedUserIds
     * @return
     */
    private List<Long> matchUserByConnectRate(Integer userSex, List<Long> excludedUserIds, int matchNum) {
        log.info("接通率匹配用户");
        return familySkillDAO.selectUserIdsByConnectRate(userSex, excludedUserIds, matchNum);
    }

    /**
     * 根据活跃度匹配
     *
     * @param userSex
     * @param excludedUserIds
     * @param matchNum
     * @return
     */
    private List<Long> matchUserByActive(Integer userSex, List<Long> excludedUserIds, int matchNum) {
        log.info("根据活跃度匹配");
        return familySkillDAO.selectUserIdsByActive(userSex, excludedUserIds, matchNum);
    }

    /**
     * 获取匹配用户头像
     *
     * @param userSex
     * @param excludedUserIds
     * @return
     */
    public List<UserMatchResponse> getMatchUserAvatarList(Integer userSex, List<Long> excludedUserIds) {
        return familySkillDAO.selectUserAvatar(userSex, VoiceChatStatus.IDLE, excludedUserIds);
    }

    /**
     * 获得匹配到的陪陪
     *
     * @param userSex  0:未知; 1:男; 2:女
     * @param maxCount 显示的最大数量
     *                 1. 所有陪陪，按登录时间由近及远取50条数据
     *                 2. 随机呈现12条数据
     * @return
     */
    public List<UserPeiPeiVO> getMatchPeipeiList(Integer userSex, int maxCount) {
        List<UserPeipeiDO> list = familySkillDAO.selectAllPeipei(userSex);
        if (CollectionUtils.isEmpty(list)) {
            return Collections.EMPTY_LIST;
        }

        list = rtcRoomService.setRtcRoomInfo(list);
        long random = System.currentTimeMillis() - 1610000000000L;
        Map<Long, UserPeipeiDO> map = list.stream()
                .collect(Collectors.toMap(p -> p.getUserId() + random, p -> p));

        int size = list.size() > maxCount ? maxCount : list.size();
        List<UserPeiPeiVO> result = new ArrayList<>(size);
        AtomicInteger i = new AtomicInteger();
        map.forEach((k, v) -> {
            if (i.getAndIncrement() >= size) {
                return;
            }
            UserPeiPeiVO vo = new UserPeiPeiVO();
            BeanUtils.copyProperties(v, vo);
            result.add(vo);
        });
        return result;
    }

    /**
     * 获取排除的用户
     *
     * @param hisId
     * @return
     */
    public List<Long> getExcludedUserIds(Long hisId) {
        return redisService.hashKeys(RedisCacheKeys.getVoiceChatExcludedUser(hisId)).stream().map(Convert::toLong).collect(Collectors.toList());
    }

    /**
     * 添加排除的用户
     *
     * @param hisId
     * @return
     */
    public boolean addExcludedUserIds(Long hisId, Long targetUserId) {
        return redisService.hashPutIfAbsent(RedisCacheKeys.getVoiceChatExcludedUser(hisId), Convert.toStr(targetUserId), Convert.toStr(System.currentTimeMillis()));
    }

    /**
     * 获取当前轮次
     *
     * @param hisId
     * @return
     */
    public int getCurRound(Long hisId) {
        return Convert.toInt(redisService.hashSize(RedisChatKeys.getVoiceChatRoundUser(hisId)), 1);
    }

    /**
     * 添加轮次用户
     *
     * @param hisId
     * @param round
     * @param matchUserIds
     */
    public void addRoundUsers(Long hisId, Integer round, List<Long> matchUserIds) {
        redisService.hashPut(RedisChatKeys.getVoiceChatRoundUser(hisId), Convert.toStr(round), JSON.toJSONString(matchUserIds));
    }


    /**
     * 添加轮次用户
     *
     * @param hisId
     * @param round
     */
    public List<Long> getRoundUserIds(Long hisId, Integer round) {
        Object o = redisService.hashGet(RedisChatKeys.getVoiceChatRoundUser(hisId), Convert.toStr(round));
        if (o != null) {
            List<Long> list = JSON.parseArray((String) o, Long.class);
            log.info("hisId {} , round {} , list {} ", hisId, round, list);
            return list;
        }
        return Collections.EMPTY_LIST;
    }
}
