package com.starbuds.server.service.media.provider;

import cn.hutool.core.convert.Convert;
import cn.hutool.http.HttpRequest;
import cn.hutool.http.HttpResponse;
import cn.hutool.http.HttpUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.rtc.RtcProviderEnum;
import com.starbuds.server.common.pojo.db.VoiceChatHisDO;
import com.starbuds.server.common.pojo.rtc.RtcAuthInfo;
import com.starbuds.server.common.pojo.rtc.RtcProviderSignature;
import com.starbuds.server.common.service.media.ProviderInterface;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.VoiceChatHisDAO;
import com.starbuds.server.service.media.zego.ZegoConfig;
import com.starbuds.server.service.media.zego.ZegoServerAssistant;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import java.security.SecureRandom;
import java.util.*;

@Slf4j
@Component
public class ZegoProvider implements ProviderInterface {

    @Resource
    private ProviderBuilder providerBuilder;

    @Resource
    VoiceChatHisDAO voiceChatHisDAO;

    @Resource
    ZegoConfig zegoConfig;

    @Reference(version = "1.0.0")
    UserService userService;

    @PostConstruct
    public void init() {
        providerBuilder.regist(RtcProviderEnum.ZEGO, this);
    }

    @Override
    public RtcAuthInfo getAuthInfo(Long roomId, Long userId) {

        Long userNo = userService.getUserNo(userId);
        if (userNo != null && userNo > 0) {
            String channelName = String.format("YL-%d%d", roomId, userId);
            RtcAuthInfo rtcAuthInfo = new RtcAuthInfo();
            rtcAuthInfo.setChannelName(channelName);
            rtcAuthInfo.setToken(buildToken(channelName, userId, 1, System.currentTimeMillis() + TimeConstants.DAY_MILLS));
            rtcAuthInfo.setUserNo(Convert.toInt(userNo));
            return rtcAuthInfo;
        }
        return null;
    }

    @Override
    public String voiceChatChannelName(Long roomId, Long userId) {
        if (roomId == null || userId == null) {
            return "";
        }

        return "" + roomId + userId;
    }

    @Override
    public String buildToken(String channelName, Long userId, int role, long expireTime) {
        ZegoServerAssistant.Privileges privileges = new ZegoServerAssistant.Privileges();
        privileges.canLoginRoom = true;
        privileges.canPublishStream = true;

        ZegoServerAssistant.TokenInfo tokenInfo = ZegoServerAssistant.generateToken(zegoConfig.appId(), zegoConfig.serverSecret(), channelName, userId.toString(), privileges, expireTime);
        return tokenInfo.data;
    }

    @Override
    public boolean disableChannel(String channelName) {
        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(Long.valueOf(channelName));
        if (hisDO == null) {
            return false;
        }

        Map<String, Object> param = new HashMap<>();
        param.put("RoomId", channelName);

        String detach = String.format("UserId[]=%d&UserId[]=%d", hisDO.getUserId(), hisDO.getTargetUserId());
        String requestPath = String.format("?Action=KickoutUser&%s", detach);
        JSONObject jsonObject = zegoNetworking(requestPath, param);
        if (jsonObject == null) {
            return true;
        }

        return jsonObject.getInteger("Code").compareTo(0) == 0;
    }

    @Override
    public boolean isInChannel(String channelName, Long userId) {
        log.info("--ZegoProvider : isInChannel");
        Map<String, Object> param = new HashMap<>(1);
        param.put("RoomId", channelName);

        JSONObject jsonObject = zegoNetworking("?Action=DescribeSimpleStreamList", param);

        if (jsonObject == null) {
            return false;
        }

        if (jsonObject.getInteger("Code").compareTo(0) == 0) {
            JSONObject dataJSON = jsonObject.getJSONObject("Data");
            JSONArray streamArray = dataJSON.getJSONArray("StreamList");
            if (streamArray != null) {
                for (Object obj : streamArray) {
                    JSONObject objJSON = (JSONObject) obj;
                    //流列表中的用户ID
                    String streamUserId = objJSON.getString("UserId");
                    if (streamUserId.equals(userId.toString())) {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    @Override
    public String startRecord(String channelName) {
        return null;
    }

    @Override
    public String stopRecord(String channelName, String resourceId, String sid) {
        return null;
    }

    @Override
    public boolean verifyNotifySign(RtcProviderSignature signature) {
        String sign = signature.getSignature();
        String nonce = signature.getNonce();
        long timestamp = signature.getTimestamp().longValue();

        String[] tempArr = {zegoConfig.callbackSecret(), "" + timestamp, nonce};
        Arrays.sort(tempArr);

        String tmpStr = "";
        for (int i = 0; i < tempArr.length; i++) {
            tmpStr += tempArr[i];
        }
        tmpStr = org.apache.commons.codec.digest.DigestUtils.sha1Hex(tmpStr);

        return tmpStr.equals(sign);
    }

    protected JSONObject zegoNetworking(String protocal, Map<String, Object> param) {
        log.info("--ZegoProvider Networking:" + protocal + "  \n\n param:" + param);
        //生成16进制随机字符串(16位)
        byte[] bytes = new byte[8];
        SecureRandom sr = null;
        //使用SecureRandom获取高强度安全随机数生成器
        try {
            sr = SecureRandom.getInstanceStrong();
        } catch (Exception e) {
            log.error("-- zegoNetworking:SecureRandom:" + e.toString());
        }
        sr.nextBytes(bytes);

        long timestamp = System.currentTimeMillis() / 1000;

        String signatureNonce = ZegoServerAssistant.bytesToHex(bytes);
        String signature = ZegoServerAssistant.generateSignature(zegoConfig.appId(), zegoConfig.serverSecret(), signatureNonce, timestamp);

        List<String> params = new ArrayList<>();
        params.add(String.format("%s=%s", "AppId", zegoConfig.appId().toString()));
        params.add(String.format("%s=%s", "Signature", signature));
        params.add(String.format("%s=%s", "SignatureNonce", signatureNonce));
        params.add(String.format("%s=%s", "SignatureVersion", "2.0"));
        params.add(String.format("%s=%d", "Timestamp", timestamp));
        params.add(String.format("%s=%s", "IsTest", zegoConfig.isTest()));

        for (String key : param.keySet()) {
            params.add(String.format("%s=%s", key, param.get(key)));
        }

        String requestParam = String.join("&", params);
        String requestURL = String.format("%s/%s&%s", zegoConfig.baseURL(), protocal, requestParam);

        // return hutoolGet(requestURL);
        return okhttpGet(requestURL);
    }

    private JSONObject hutoolGet(String requestURL) {
        JSONObject jsonObj = null;
        try {
            log.info("--ZegoProvider requestURL:" + requestURL);
            String rspData = HttpUtil.get(requestURL, 1000 * 30);
            log.info("--ZegoProvider Networking:requestURL:" + requestURL + "  \n\n response:" + rspData);
            if (StringUtils.isNotBlank(rspData)) {
                jsonObj = JSON.parseObject(rspData);
            }
        } catch (Exception e) {
            log.error("调用即构URL:" + requestURL + "    失败！！");
        }

        return jsonObj;
    }

    private JSONObject okhttpGet(String requestURL) {
        JSONObject jsonObj = null;
        try {
            log.info("--ZegoProvider requestURL:" + requestURL);
            String rspData = com.starbuds.server.service.media.util.HttpUtil.get(requestURL);
            log.info("--ZegoProvider Networking:requestURL:" + requestURL + "  \n\n response:" + rspData);
            if (StringUtils.isNotBlank(rspData)) {
                jsonObj = JSON.parseObject(rspData);
            }
        } catch (Exception e) {
            log.error("调用即构URL:" + requestURL + "    失败！！");
        }

        return jsonObj;
    }

}
