package com.starbuds.server.service.media.rocket.handler;

import cn.hutool.core.convert.Convert;
import com.starbuds.server.common.constant.cache.RedisCacheKeys;
import com.starbuds.server.common.constant.cache.RedisChatKeys;
import com.starbuds.server.common.constant.cache.RedisFinanceKeys;
import com.starbuds.server.common.constant.common.TimeConstants;
import com.starbuds.server.common.constant.finance.ChatTimeWalletType;
import com.starbuds.server.common.constant.finance.WalletTypeEnum;
import com.starbuds.server.common.constant.media.*;
import com.starbuds.server.common.constant.message.im.ImTargetType;
import com.starbuds.server.common.constant.rocketmq.RocketTag;
import com.starbuds.server.common.constant.rocketmq.RocketTopic;
import com.starbuds.server.common.constant.statistics.StatPlatformTypeEnum;
import com.starbuds.server.common.constant.user.SingleChatStatus;
import com.starbuds.server.common.constant.user.VoiceChatStatus;
import com.starbuds.server.common.pojo.db.LiveHisDO;
import com.starbuds.server.common.pojo.db.SingleChatHisDO;
import com.starbuds.server.common.pojo.db.VoiceChatHisDO;
import com.starbuds.server.common.pojo.message.im.chat.*;
import com.starbuds.server.common.pojo.mq.*;
import com.starbuds.server.common.pojo.user.UserCardVO;
import com.starbuds.server.common.service.data.RedisService;
import com.starbuds.server.common.service.finance.PayService;
import com.starbuds.server.common.service.finance.WalletService;
import com.starbuds.server.common.service.media.LiveService;
import com.starbuds.server.common.service.media.VoiceChatService;
import com.starbuds.server.common.service.message.ImService;
import com.starbuds.server.common.service.statistics.StatPlatformService;
import com.starbuds.server.common.service.user.UserService;
import com.starbuds.server.service.media.dao.*;
import com.starbuds.server.service.media.provider.*;
import com.starbuds.server.service.media.rocket.RocketProducer;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.math.NumberUtils;
import org.apache.dubbo.config.annotation.Reference;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;

/**
 * 备注
 *
 * @author Lay
 * @date 2019/12/25
 */
@Component
@Slf4j
public class DelayMsgHandler {

    @Reference(version = "1.0.0")
    LiveService liveService;

    @Reference(version = "1.0.0")
    ImService imService;

    @Reference(version = "1.0.0")
    PayService payService;

    @Reference(version = "1.0.0")
    WalletService walletService;

    @Reference(version = "1.0.0")
    StatPlatformService statPlatformService;

    @Reference(version = "1.0.0")
    RedisService redisService;

    @Reference(version = "1.0.0")
    UserService userService;

    @Resource
    LiveHeatProvider liveHeatProvider;

    @Resource
    SingleChatProvider singleChatProvider;

    @Resource
    ChatMatchProvider chatMatchProvider;

    @Resource
    ProviderBuilder providerBuilder;

    @Resource
    RocketProducer rocketProducer;

    @Resource
    BattleProvider battleProvider;

    @Resource
    LiveHisDAO liveHisDAO;

    @Resource
    LiveStreamHisDAO liveStreamHisDAO;

    @Resource
    SingleChatHisDAO singleChatHisDAO;

    @Resource
    AnchorDAO anchorDAO;

    @Resource
    VoiceChatHisDAO voiceChatHisDAO;

    @Resource
    VoiceChatProvider voiceChatProvider;

    @Resource
    VoiceChatService voiceChatService;

    /**
     * 直播热度过期处理
     *
     * @param msg
     */
    public void handleLiveHeatExpire(QueueLiveHeatExpireMsg msg) {
        if (msg.getHeatKey() != null && msg.getHeatKey().equals(liveHeatProvider.getHeatKey(msg.getUserId()))) {
            // 减热度
            liveHeatProvider.changeLiveHeat(msg.getUserId(), -msg.getHeatValue(), 0L);
        }
    }

    /**
     * 直播超时处理
     *
     * @param msg
     */
    public void handleLiveTimeout(QueueLiveTimeoutMsg msg) {
        log.info("直播超时检查 liveHis:{} streamHis:{}", msg.getLiveHisId(), msg.getStreamHisId());
        // 判断是否超时 1 直播状态是否还是直播中 2 推流是否有更新
        LiveHisDO liveHisDO = liveHisDAO.selectById(msg.getLiveHisId());
        if (LiveStatusEnum.Living == LiveStatusEnum.toEnum(liveHisDO.getLiveStatus())) {
            // 推流是否更新
            if (msg.getStreamHisId().equals(liveStreamHisDAO.selectLastIdByLiveHisId(msg.getLiveHisId()))) {
                // 强制下播
                liveService.stopLive(msg.getLiveHisId(), true);
            }
        }
    }

    /**
     * 单聊呼叫超时
     *
     * @param msg
     */
    @Transactional(rollbackFor = Exception.class)
    public void handleSingleChatRingTimeout(QueueSingleChatRingTimeoutMsg msg) {
        SingleChatHisDO singleChatHis = singleChatHisDAO.selectById(msg.getHisId());
        if (singleChatHis != null) {
            ChatStatusEnum chatStatus = ChatStatusEnum.toEnum(singleChatHis.getHisStatus());
            if (ChatStatusEnum.Ringing == chatStatus) {
                log.info("单聊呼叫超时 {}", msg.getHisId());
                if (singleChatHisDAO.updateStatus(msg.getHisId(), ChatStatusEnum.Canceled.getValue(), ChatStatusEnum.Ringing.getValue(), System.currentTimeMillis()) == 1) {
                    // 更新主播状态
                    anchorDAO.updateSingleChatStatus(singleChatHis.getTargetUserId(), SingleChatStatus.IDLE, SingleChatStatus.BUSY);

                    // 发送超时消息
                    SingleChatRingTimeoutMsg ringTimeoutMsg = new SingleChatRingTimeoutMsg();
                    ringTimeoutMsg.setHisId(msg.getHisId());
                    imService.sendCustomImMsg(null, singleChatHis.getUserId(), ringTimeoutMsg, null);
                    imService.sendCustomImMsg(null, singleChatHis.getTargetUserId(), ringTimeoutMsg, null);

                    // 匹配超时不计入主播超时
                    if (!msg.getIsMatch()) {
                        // 主播超时次数
                        anchorDAO.increaseSingleChatTimeoutTimes(singleChatHis.getTargetUserId(), 1);
                        statPlatformService.addPlatformStat(singleChatHis.getChannelId(), StatPlatformTypeEnum.SINGLE_CHAT_TIMES_TIMEOUT, BigDecimal.ONE);
                    } else {
                        statPlatformService.addPlatformStat(singleChatHis.getChannelId(), StatPlatformTypeEnum.SINGLE_CHAT_TIMES_MATCH_TIMEOUT, BigDecimal.ONE);
                    }

                    // 将主播加回池子
                    chatMatchProvider.addMatchPoolAnchor(singleChatHis.getTargetUserId());

                    // 解除锁定
                    redisService.delKey(RedisCacheKeys.getSingleChatMatchLock(singleChatHis.getUserId()));
                    redisService.delKey(RedisCacheKeys.getSingleChatMatchLock(singleChatHis.getTargetUserId()));
                }
            }
        }
    }

    /**
     * 单聊计费周期
     *
     * @param msg
     */
    public void handleSingleChatChargingTick(QueueSingleChatChargingTickMsg msg) {
        log.info("单聊计费周期 {} {}", msg.getHisId(), msg.getTimes());
        SingleChatHisDO singleChatHis = singleChatHisDAO.selectById(msg.getHisId());
        if (singleChatHis == null || ChatStatusEnum.Calling != ChatStatusEnum.toEnum(singleChatHis.getHisStatus())) {
            return;
        }

        // 频道判断
        if (!providerBuilder.provider(msg.getAppVer()).isInChannel(singleChatHis.getHisId(), singleChatHis.getUserId())) {
            singleChatProvider.stopChat(singleChatHis.getHisId(), SingleChatStopReasonEnum.UserClose, msg.getAppVer());
            return;
        }

        Long timestamp = System.currentTimeMillis();
        Long balance = payService.payForSingleChat(singleChatHis.getHisId(), msg.getTimes(), timestamp);
        if (balance == null || balance < 0) {
            singleChatProvider.stopChat(singleChatHis.getHisId(), balance == null ? SingleChatStopReasonEnum.ChargingFailed : SingleChatStopReasonEnum.BalanceNotEnough, msg.getAppVer());
            return;
        }

        // 发送下一计费周期
        msg.setTimes(msg.getTimes() + 1);
        msg.setTimestamp(timestamp);
        rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.SINGLE_CHAT_CHARGING_TICK, msg.serialize(), timestamp + TimeConstants.MIN_MILLS);

        long minutesRemain = balance / singleChatHis.getHisPrice();
        minutesRemain += walletService.getChatTimeWalletMinutes(singleChatHis.getUserId(), ChatTimeWalletType.SINGLE_CHAT);
        if (minutesRemain <= 5) {
            // 发送余额提醒
            SingleChatBalanceRemindMsg singleChatBalanceRemindMsg = new SingleChatBalanceRemindMsg();
            singleChatBalanceRemindMsg.setHisId(singleChatHis.getHisId());
            singleChatBalanceRemindMsg.setMinutesRemain(0L);
            imService.sendChatGroupMsg(ImTargetType.GROUP, null, null, singleChatHis.getHisId(), singleChatBalanceRemindMsg);
        }
    }

    /**
     * 语音速配计费周期
     *
     * @param msg
     */
    public void handleVoiceChatChargingTick(QueueVoiceChatChargingTickMsg msg) {
        log.info("语音速配计费周期 {}", msg.getHisId(), msg.getTimes());

        Long hisId = msg.getHisId();
        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);
        ChatStatusEnum statusEnum = ChatStatusEnum.toEnum(hisDO.getHisStatus());
        if (hisDO == null) {
            log.info("未找到记录{}", hisId);
            return;
        }

        if (ChatStatusEnum.Calling != statusEnum) {
            log.info("{}当前状态{}，不支持计费", hisId, statusEnum.getDesc());
            return;
        }

        Long userId = hisDO.getUserId();
        long timestamp = msg.getTimestamp();
        Long targetUserId = msg.getTargetUserId();

        // 频道判断
        if (!providerBuilder.provider(msg.getAppVer()).isInChannel(Convert.toStr(hisId), userId)) {
            log.info("用户离开频道{}", hisId);
            voiceChatService.stopChat(hisId, VoiceChatStopReasonEnum.UserLeave, timestamp, 0, msg.getAppVer());
            return;
        }

        boolean useFreeTime = msg.isUseFreeTime();
        // 要计那次的费用 = 当前要计那一次的费用
        Integer times = msg.getTimes();

        log.info("语音速配{}第{}次计费 使用免费时长{}", hisId, times, times > 1 ? false : useFreeTime);

        // 使用余额
        if (!useFreeTime) {
            // 最近一次已经计费
            Integer lastChargingTimes = Convert.toInt(redisService.hashGet(RedisFinanceKeys.VOICE_CHAT_CHARGING, Convert.toStr(hisId)));
            if (lastChargingTimes == null) {
                log.info("累计异常{}", hisId);
                return;
            }
            // 已经计费的次数大于本次
            if (lastChargingTimes >= times) {
                // 该次已计费
                log.info("hisId {} 已计费,lastChargingTimes {},times {}", hisId, lastChargingTimes, times);
                return;
            }

            log.info("正在结算 hisId {},times {}", hisId, times);
            Long balance = payService.payForVoiceChat(hisId, userId, targetUserId, hisDO.getChannelId(), hisDO.getHisPrice());
            if (balance == null || balance < 0) {
                log.info("计费失败或余额不足 hisId {},balance {}", hisId, balance);
                voiceChatService.stopChat(hisId, balance == null ? VoiceChatStopReasonEnum.ChargingFailed : VoiceChatStopReasonEnum.BalanceNotEnough, timestamp, 0, msg.getAppVer());
                return;
            }
            log.info("结算成功,hisId {},times {}", hisId, times);
        }

        long startDeliverTime = timestamp + TimeConstants.MIN_MILLS;

        // 发送下一计费周期
        QueueVoiceChatChargingTickMsg tickMsg = new QueueVoiceChatChargingTickMsg();
        tickMsg.setHisId(hisId);
        tickMsg.setTargetUserId(targetUserId);
        tickMsg.setTimes(times + 1);
        tickMsg.setUseFreeTime(false);
        tickMsg.setTimestamp(startDeliverTime);
        tickMsg.setAppVer(msg.getAppVer());
        rocketProducer.sendMsg(RocketTopic.TPC_DELAY, RocketTag.VOICE_CHAT_CHARGING_TICK, tickMsg.serialize(), startDeliverTime);

        // 少于三分钟
        Long userBalance = walletService.getBalance(userId, WalletTypeEnum.STAR_DIAMOND);
        if (userBalance < 3 * hisDO.getHisPrice()) {
            // 发送余额提醒
            VoiceChatBalanceRemindMsg remindMsg = new VoiceChatBalanceRemindMsg();
            remindMsg.setHisId(hisId);
            remindMsg.setMinutesRemain(userBalance / hisDO.getHisPrice());
            imService.sendChatGroupMsg(ImTargetType.GROUP, null, null, Convert.toStr(hisId), remindMsg);
        }
    }

    /**
     * 处理单聊匹配延迟
     *
     * @param msg
     */
    public void handleSingleChatMatchDelay(QueueSingleChatMatchDelayMsg msg) {
        chatMatchProvider.matchUser(msg.getUserId());
    }

    /**
     * PK 邀请超时
     *
     * @param msg
     */
    public void handleBattleInviteTimeout(QueueBattleInviteTimeoutMsg msg) {
        log.info("PK 邀请超时 {}", msg.getBattleId());
        battleProvider.battleInviteTimeout(msg.getBattleId(), msg.getUserId(), msg.getDestUserId());
    }

    /**
     * PK 等待超时
     *
     * @param msg
     */
    public void handleBattleWaitTimeout(QueueBattleWaitTimeoutMsg msg) {
        log.info("PK 等待超时 {}", msg.getBattleId());
        // 判断 pk 状态
        Integer battlingStatus = battleProvider.getBattlingStatus(msg.getBattleId());
        if (battlingStatus != null && BattlingStatus.Wait == battlingStatus) {
            // 超时结束
            battleProvider.battleOver(msg.getBattleId());
        }
    }

    /**
     * PK 阶段改变
     *
     * @param msg
     */
    public void handleBattleStageDelay(QueueBattleStageDelayMsg msg) {
        log.info("PK 阶段改变 {} {}", msg.getBattleId(), msg.getStage());
        if (BattlingStatus.Punish == msg.getStage()) {
            // 结算进入惩罚
            battleProvider.battleSettle(msg.getBattleId());
        } else if (BattlingStatus.Over == msg.getStage()) {
            // 结束 PK
            battleProvider.battleOver(msg.getBattleId());
        }
    }

    /**
     * 语音速配超时
     *
     * @param msg
     */
    public void handleVoiceMatchTimeout(QueueVoiceMatchTimeoutMsg msg) {
        log.info("语音速配超时 {}", msg.getHisId());

        Long hisId = msg.getHisId();
        long timestamp = msg.getTimestamp();

        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);
        if (hisDO == null) {
            return;
        }

        Long userId = hisDO.getUserId();
        Long channelId = hisDO.getChannelId();
        ChatStatusEnum statusEnum = ChatStatusEnum.toEnum(hisDO.getHisStatus());
        Long targetUserId = hisDO.getTargetUserId();

        if (statusEnum == ChatStatusEnum.Ringing) {
            log.info("已超时 {}", msg.getHisId());
            // 统计记录语音速配匹配时间
            long matchStartTime = NumberUtils.toLong(redisService.get(RedisChatKeys.getVoiceChatMatchStartTime(userId)), 0);
            if (matchStartTime > 0) {
                long matchDuration = System.currentTimeMillis() - matchStartTime;
                statPlatformService.addPlatformStat(hisDO.getChannelId(), StatPlatformTypeEnum.VOICE_CHAT_MATCH_DURATION, new BigDecimal("" + matchDuration));
            }

            // 更新到数据库
            VoiceChatHisDO hisUpdate = new VoiceChatHisDO();
            hisUpdate.setHisId(hisId);
            hisUpdate.setHisStatus(ChatStatusEnum.Canceled.getValue());
            hisUpdate.setHisUpdateTime(timestamp);
            hisUpdate.setMatchEndTime(timestamp);
            int updateCount = voiceChatHisDAO.updateByIdAndWhereStatus(hisUpdate, ChatStatusEnum.Ringing.getValue());
            if (updateCount == 0) {
                log.info("{}状态已发生变化。", hisId);
                return;
            }

            // 更新用户聊天状态
            userService.updateVoiceChatStatus(userId, VoiceChatStatus.IDLE);
            userService.updateVoiceChatStatus(targetUserId, VoiceChatStatus.IDLE);

            // 速配失败
            VoiceChatMatchFailMsg matchMsgToUser = new VoiceChatMatchFailMsg();
            matchMsgToUser.setHisId(hisId);
            matchMsgToUser.setUserId(userId);
            matchMsgToUser.setTargetUserId(targetUserId);
            matchMsgToUser.setMsg(targetUserId == null ? "未匹配到用户" : "对方未同意");
            matchMsgToUser.setMatchStartTime(hisDO.getMatchStartTime());
            matchMsgToUser.setMatchEndTime(timestamp);

            UserCardVO userCard = userService.getCard(userId);
            matchMsgToUser.setUserCard(userCard);

            UserCardVO targetUserCard = null;
            if (targetUserId != null) {
                targetUserCard = userService.getCard(targetUserId);
                matchMsgToUser.setTargetUserCard(targetUserCard);
            }

            imService.sendCustomImMsg(null, userId, matchMsgToUser, null);

            if (targetUserId != null) {
                VoiceChatMatchFailMsg matchMsgToTargetUser = new VoiceChatMatchFailMsg();
                matchMsgToTargetUser.setHisId(hisId);
                matchMsgToTargetUser.setUserId(userId);
                matchMsgToTargetUser.setTargetUserId(targetUserId);
                matchMsgToTargetUser.setMsg(targetUserId == null ? "未匹配到用户" : "对方未同意");
                matchMsgToTargetUser.setMatchStartTime(hisDO.getMatchStartTime());
                matchMsgToTargetUser.setMatchEndTime(timestamp);
                matchMsgToTargetUser.setUserCard(userCard);
                matchMsgToTargetUser.setTargetUserCard(targetUserCard);
                imService.sendCustomImMsg(null, targetUserId, matchMsgToTargetUser, null);
            }

            // 统计
            statPlatformService.addPlatformStat(channelId, StatPlatformTypeEnum.VOICE_CHAT_TIMES_MATCH_TIMEOUT, BigDecimal.ONE);
        } else {
            log.info("语音速配{}{},不作超时处理", hisId, statusEnum.getDesc());
        }

        return;
    }

    /**
     * 语音速配等待超时
     *
     * @param msg
     */
    public void handleVoiceMatchWaitTimeout(QueueVoiceMatchWaitTimeoutMsg msg) {
        log.info("语音速配等待超时");
        Long hisId = msg.getHisId();
        VoiceChatHisDO hisDO = voiceChatHisDAO.selectById(hisId);
        if (hisDO == null) {
            return;
        }
        // 如果不是当前目标用户接通
        if (!msg.getCurTargetUserId().equals(hisDO.getTargetUserId())) {
            log.info("10秒内未处理，被系统取消, hisId {} curTargetUserId {} round {}", msg.getHisId(), msg.getCurTargetUserId(), msg.getRound());
            // 更新语音状态
            userService.updateVoiceChatStatus(msg.getCurTargetUserId(), VoiceChatStatus.IDLE);

            // 拒绝mq
            QueueVoiceChatRejectMsg rejectMsg = new QueueVoiceChatRejectMsg();
            rejectMsg.setHisId(hisId);
            rejectMsg.setUserId(hisDO.getUserId());
            rejectMsg.setTargetUserId(msg.getCurTargetUserId());
            rejectMsg.setRound(msg.getRound());
            rejectMsg.setMatchStartTime(msg.getMatchStartTime());
            rejectMsg.setMatchShouldEndTime(msg.getMatchStartTime() + TimeConstants.MIN_MILLS);
            rejectMsg.setExcludedUserIds(voiceChatProvider.getExcludedUserIds(hisId));
            rocketProducer.sendMsg(RocketTopic.TPC_NORMAL, RocketTag.VOICE_CHAT_REJECT, rejectMsg.serialize());


            // 取消消息发给对方
            VoiceChatCancelMsg cancelMsgToTargetUser = new VoiceChatCancelMsg();
            cancelMsgToTargetUser.setHisId(hisId);
            cancelMsgToTargetUser.setUserId(hisDO.getUserId());
            cancelMsgToTargetUser.setTargetUserId(msg.getCurTargetUserId());
            cancelMsgToTargetUser.setHisCancelTime(msg.getTimestamp());
            cancelMsgToTargetUser.setMsg("10秒内未处理，被系统取消");
            imService.sendCustomImMsg(null, msg.getCurTargetUserId(), cancelMsgToTargetUser, null);
        }
    }
}
